# Copyright 2016 - 2019  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only


async def test_comment(client):
    await client.authenticate('test', 'test_pw')

    res = await client.post('/marv/api/comment', headers=client.headers, json={})
    assert res.status == 400

    res = await client.post_json('/marv/api/comment', json={
        '1': {
            'add': ['lorem ipsum', 'dolor'],
        },
    })
    assert res == {}



# Copyright 2016 - 2019  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

import asyncio
import base64
import os
from collections import namedtuple
from logging import getLogger
from pathlib import Path

from aiohttp import web

import marv_webapi
from marv.collection import cached_property
from marv_webapi.tooling import safejoin


LOADED = False
log = getLogger(__name__)


async def site_load_for_web(aioapp):
    aioapp['site'].load_for_web()


async def destroy_site(aioapp):
    await aioapp['site'].destroy()


class App():
    STARTUP_FNS = (
        site_load_for_web,
    )

    SHUTDOWN_FNS = (
        destroy_site,
    )

    NOCACHE = {'Cache-Control': 'no-cache'}

    def __init__(self, site, app_root='', middlewares=None):
        self.aioapp = web.Application(middlewares=middlewares or [])
        self.aioapp['api_endpoints'] = {}
        self.aioapp['app_root'] = app_root.rstrip('/')
        self.aioapp['config'] = {
            'SECRET_KEY': Path(site.config.marv.sessionkey_file).read_text(),
        }
        self.aioapp['debug'] = False
        self.aioapp['route_acl'] = site.config.marv.acl()
        self.aioapp['site'] = site
        self.aioapp.route = self.route

        for func in self.STARTUP_FNS:
            self.aioapp.on_startup.append(func)

        for func in self.SHUTDOWN_FNS:
            self.aioapp.on_shutdown.append(func)

        self.initialize_routes()

    def route(self, path):
        def dec(func):
            self.aioapp.add_routes([web.route('GET', f'{self.aioapp["app_root"]}{path}', func)])
        return dec

    @cached_property
    def index_html(self):
        path = Path(self.aioapp['site'].config.marv.staticdir, 'index.html')
        index_html = path.read_text().replace('MARV_APP_ROOT', self.aioapp['app_root'] or '')

        frontenddir = Path(self.aioapp['site'].config.marv.frontenddir)
        for ext in ('css', 'js'):
            try:
                data = base64.b64encode((frontenddir / f'custom.{ext}').read_bytes()).decode()
            except IOError:
                pass
            else:
                placeholder = f'<!--custom.{ext}-->'
                assert placeholder in index_html
                script = f'<script src="data:text/javascript;base64,{data}"></script>' \
                         if ext == 'js' else \
                         f'<link rel="stylesheet" href="data:text/css;base64,{data}" />'
                index_html = index_html.replace(placeholder, script, 1)
        return index_html

    def initialize_routes(self):
        global LOADED  # pylint: disable=global-statement
        if not LOADED:
            marv_webapi.load_entry_points()
            LOADED = True
        marv_webapi.webapi.init_app(self.aioapp, url_prefix='/marv/api',
                                    app_root=self.aioapp['app_root'])

        customdir = Path(self.aioapp['site'].config.marv.frontenddir, 'custom')
        staticdir = Path(self.aioapp['site'].config.marv.staticdir)

        # decorator for non api endpoint routes
        @self.route('/custom/{path:.*}')
        async def custom(request):  # pylint: disable=unused-variable
            path = request.match_info['path']
            fullpath = safejoin(customdir, path)
            if not fullpath.is_file():
                raise web.HTTPNotFound
            return web.FileResponse(fullpath, headers=self.NOCACHE)

        @self.route('/{path:.*}')
        async def assets(request):  # pylint: disable=unused-variable
            path = request.match_info['path'] or 'index.html'
            if path == 'index.html':
                return web.Response(text=self.index_html, headers={
                    'Content-Type': 'text/html',
                    **self.NOCACHE,
                })

            fullpath = safejoin(staticdir, path)
            if not fullpath.is_file():
                raise web.HTTPNotFound
            return web.FileResponse(fullpath, headers=self.NOCACHE)



# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

import marv_api
from marv_api import deprecation
from marv_webapi.tooling import api_endpoint as _api_endpoint
from marv_webapi.tooling import api_group as _api_group

DEPRECATIONS = {
    'api_endpoint': deprecation.Info(__name__, '20.07', _api_endpoint),
    'api_group': deprecation.Info(__name__, '20.07', _api_group),
}

DEPRECATIONS.update(
    (name, deprecation.Info(
        __name__, '20.07', getattr(marv_api, name),
        f'Use marv_api.{name} instead: import marv_api as marv.',
    )) for name in (
        'Abort',
        'create_group',
        'create_stream',
        'get_logger',
        'get_requested',
        'input',
        'make_file',
        'node',
        'pull',
        'pull_all',
        'push',
        'select',
        'set_header',
    )
)

__all__ = ()
__dir__, __getattr__ = deprecation.dir_and_getattr(__name__, __all__, DEPRECATIONS)



# Copyright 2016 - 2019  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

import asyncio
import code
import datetime
import functools
import json
import signal
import sqlite3
import sys
from contextlib import asynccontextmanager
from functools import reduce
from logging import getLogger
from pathlib import Path

import click
from aiohttp import web
from gunicorn.app.base import BaseApplication
from gunicorn.arbiter import Arbiter
from jinja2 import Template
from tortoise.exceptions import DoesNotExist

from marv.db import DBError, DBNotInitialized, DBVersionError, USERGROUP_REGEX
from marv.site import UnknownNode, load_sitepackages, make_config
from marv.utils import echo, err, find_obj, within_pyinstaller_bundle
from marv_cli import PDB
from marv_cli import marv as marvcli
from marv_node.stream import RequestedMessageTooOld
from marv_store import DirectoryAlreadyExists

try:
    import marv_ee
    from marv_ee.app import App
    from marv_ee.site import Site
except ImportError:
    marv_ee = None
    from marv.app import App
    from marv.site import Site

log = getLogger(__name__)


class NodeParamType(click.ParamType):
    name = 'NODE'

    def convert(self, value, param, ctx):
        return find_obj(value)


NODE = NodeParamType()


def get_site_config():
    ctx = click.get_current_context()
    siteconf = ctx.obj
    if siteconf is None:
        ctx.fail('Could not find config file: ./marv.conf or /etc/marv/marv.conf.\n'
                 'Change working directory or specify explicitly:\n\n'
                 '  marv --config /path/to/marv.conf\n')
    return siteconf


@asynccontextmanager
async def create_site(init=None):
    siteconf = get_site_config()
    try:
        site = await Site.create(siteconf, init=init)
    except (sqlite3.OperationalError, DBNotInitialized) as exc:
        if PDB:
            raise
        err(f'{exc!r}\n\nDid you run marv init?\n', exit=1)
    except DBVersionError as exc:
        err(f'{exc!r}\n\n'
            'Existing database is not compatible with this version of MARV. '
            'Check the migration instructions.', exit=1)
    try:
        yield site
    finally:
        await site.destroy()


def click_async(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return asyncio.run(func(*args, **kwargs))
    return wrapper


class NoSigintArbiter(Arbiter):
    def handle_int(self):
        signal.signal(signal.SIGINT, lambda *_: self.kill_workers(signal.SIGKILL))
        super().handle_term()


class GunicornApplication(BaseApplication):  # pylint: disable=abstract-method
    def __init__(self, app_factory, bind, certfile, keyfile, *args, **kw):
        # pylint: disable=too-many-arguments
        self.app_factory = app_factory
        self.bind = bind
        self.certfile = certfile
        self.keyfile = keyfile
        super().__init__(*args, **kw)

    def load_config(self):
        self.cfg.set('proc_name', 'marvweb')
        self.cfg.set('worker_class', 'aiohttp.GunicornUVLoopWebWorker')
        self.cfg.set('bind', self.bind)
        self.cfg.set('certfile', self.certfile)
        self.cfg.set('keyfile', self.keyfile)
        self.cfg.set('workers', 1)
        self.cfg.set('graceful_timeout', 86400)

    def load(self):
        return self.app_factory

    def run(self):
        try:
            NoSigintArbiter(self).run()
        except RuntimeError as e:
            err(f'\nERROR: {e}', exit=1)


@marvcli.command('cleanup')
@click.option('--discarded/--no-discarded', help='Delete discarded datasets')
@click.option('--filters/--no-filters', help='Cleanup unused filter values')
@click.pass_context
@click_async
async def marvcli_cleanup(ctx, discarded, filters):
    """Cleanup subset filters and discarded datasets."""
    if not any([discarded, filters]):
        click.echo(ctx.get_help())
        ctx.exit(1)

    async with create_site() as site:
        if discarded:
            await site.cleanup_discarded()

        if filters:
            await site.cleanup_relations()

        # TODO: cleanup unused store paths / unused generations


@marvcli.group('develop')
def marvcli_develop():
    """Development tools."""


@marvcli_develop.command('server')
@click.option('--port', default=5000, help='Port to listen on')
@click.option('--public/--no-public', help='Listen on all IPs instead of only 127.0.0.1')
def marvcli_develop_server(port, public):
    """Run development webserver.

    ATTENTION: By default it is only served on localhost. To run it
    within a container and access it from the outside, you need to
    forward the port and tell it to listen on all IPs instead of only
    localhost.
    """
    middlewares = []
    if PDB:
        @web.middleware
        async def pdb_middleware(request, handler):
            try:
                return await handler(request)
            except Exception:  # pylint: disable=broad-except
                import pdb  # pylint: disable=import-outside-toplevel
                pdb.xpm()  # pylint: disable=no-member

        middlewares.append(pdb_middleware)

    async def create_app(middlewares=None):
        siteconf = get_site_config()
        site = await Site.create(siteconf)
        return App(site, middlewares=middlewares).aioapp

    loop = asyncio.get_event_loop()
    app = loop.run_until_complete(create_app(middlewares=middlewares))
    app['debug'] = True
    web.run_app(app,
                host=('0.0.0.0' if public else '127.0.0.1'),
                port=port)


@marvcli.command('serve')
@click.option('--host', default='127.0.0.1', help='Host to bind to')
@click.option('--port', default=8000, help='Port to listen on')
@click.option('--certfile', default=None, help='SSL certificate')
@click.option('--keyfile', default=None, help='SSL keyfile')
@click.option('--approot', default='/', help='Application root to serve', show_default=True)
def marvcli_serve(host, port, certfile, keyfile, approot):
    """Run webserver through gunicorn."""
    config = get_site_config()

    async def app_factory():
        try:
            site = await Site.create(config)
            application = App(site, app_root=approot).aioapp
        except (sqlite3.OperationalError, DBNotInitialized) as e:
            err(f'{e!r}\nDid you run marv init?', exit=1)
        except DBVersionError as e:
            err(f'{e!r}\n'
                'Existing database is not compatible with this version of MARV. '
                'Check the migration instructions.', exit=1)
        except PermissionError as e:
            err(e, exit=1)
        return application

    GunicornApplication(app_factory, f'{host}:{port}', certfile, keyfile).run()


@marvcli.command('discard')
@click.option(
    '--comments/--no-comments',
    help='Delete comments',
)
@click.option(
    '--tags/--no-tags',
    help='Delete tags',
)
@click.option(
    '--confirm/--no-confirm', default=True, show_default=True,
    help='Ask for confirmation before deleting tags and comments',
)
@click.argument('datasets', nargs=-1, required=True)
@click_async
async def marvcli_discard(datasets, tags, comments, confirm):
    """Mark DATASETS to be discarded or discard associated data.

    Without any options the specified datasets are marked to be
    discarded via `marv cleanup --discarded`. Use `marv undiscard` to
    undo this operation.

    Otherwise, selected data associated with the specified datasets is
    discarded right away.
    """
    delete_related = list(filter(None, [
        'comments' if comments else None,
        'tags' if tags else None,
    ]))
    if delete_related and confirm:
        msg = ''.join(f'  - {x}\n' for x in delete_related)
        click.confirm(f'About to PERMANENTLY delete:\n{msg}Do you want to continue?', abort=True)

    async with create_site() as site:
        if delete_related:
            await site.db.delete_comments_tags(datasets, comments, tags)
        else:
            await site.db.discard_datasets_by_setids(datasets)


@marvcli.command('undiscard')
@click.argument('datasets', nargs=-1, required=True)
@click_async
async def marvcli_undiscard(datasets):
    """Undiscard DATASETS previously discarded."""
    async with create_site() as site:
        await site.db.discard_datasets_by_setids(datasets, False)


@marvcli.command('dump')
@click.argument('dump_file', type=click.File('w'))
@click_async
async def marvcli_dump(dump_file):
    """Dump database to json file.

    Use '-' for stdout.
    """
    siteconf = make_config(get_site_config())
    dump = await Site.Database.dump_database(siteconf.marv.dburi)
    json.dump(dump, dump_file, sort_keys=True, indent=2)


@marvcli.command('restore')
@click.argument('dump_file', type=click.File())
@click_async
async def marvcli_restore(dump_file):
    """Restore previously dumped database from file.

    Use '-' to read from stdin.
    """
    data = json.load(dump_file)

    async with create_site() as site:
        await site.restore_database(**data)


@marvcli.command('init')
@click_async
async def marvcli_init():
    """(Re-)initialize marv site according to config."""
    async with create_site(init=True):
        pass


@marvcli.command('query')
@click.option(
    '--list-tags', is_flag=True,
    help='List all tags',
)
@click.option(
    '--col', '--collection', 'collections', multiple=True,
    help='Limit to one or more collections or force listing of all with --collection=*',
)
@click.option(
    '--discarded/--no-discarded',
    help='Dataset is discarded',
)
@click.option(
    '--missing', is_flag=True,
    help='Datasets with missing files',
)
@click.option(
    '--outdated', is_flag=True,
    help='Datasets with outdated node output',
)
@click.option(
    '--path', type=click.Path(resolve_path=True),
    help='Dataset contains files whose path starts with PATH',
)
@click.option(
    '--tagged', 'tags', multiple=True,
    help='Match any given tag',
)
@click.option(
    '-0', '--null', is_flag=True,
    help='Use null byte to separate output instead of newlines',
)
@click.pass_context
@click_async
async def marvcli_query(ctx, list_tags, collections, discarded,
                        missing, outdated, path, tags, null):
    """Query datasets.

    Use --col=* to list all datasets across all collections.
    """
    # pylint: disable=too-many-arguments

    if not any([collections, discarded, list_tags, missing, outdated, path, tags]):
        click.echo(ctx.get_help())
        ctx.exit(1)

    sep = '\x00' if null else '\n'

    async with create_site() as site:
        if '*' in collections:
            collections = None
        else:
            for col in collections:
                if col not in site.collections:
                    ctx.fail(f'Unknown collection: {col}')

        if list_tags:
            tags = await site.db.list_tags(collections)
            if tags:
                click.echo(sep.join(tags), nl=not null)
            return

        setids = await site.db.query(collections, discarded, outdated, path, tags, missing=missing)
        if setids:
            sep = '\x00' if null else '\n'
            click.echo(sep.join(str(x) for x in setids), nl=not null)


@marvcli.command('run', short_help='Run nodes for DATASETS')  # noqa: C901
@click.option(
    '--node', 'selected_nodes', multiple=True,
    help=(
        'Run individual nodes instead of all nodes used by detail and listing '
        'Use --list-nodes for a list of known nodes. Beyond that any node can be run '
        'by referencing it with package.module:node'
    ),
)
@click.option(
    '--list-nodes', is_flag=True,
    help='List known nodes instead of running',
)
@click.option(
    '--list-dependent', is_flag=True,
    help='List nodes depending on selected nodes',
)
@click.option(
    '--deps/--no-deps', default=True, show_default=True,
    help='Run dependencies of requested nodes',
)
@click.option(
    '--exclude', 'excluded_nodes', multiple=True,
    help='Exclude individual nodes instead of running all nodes used by detail and listing',
)
@click.option(
    '-f', '--force/--no-force',
    help='Force run of nodes whose output is already available from store',
)
# TODO: force-deps is bad as it does not allow to resume
# better: discard nodes selectively or discard --deps node
# discarding does not delete, but simply registers the most recent generation to be empty
@click.option(
    '--force-deps/--no-force-deps',
    help='Force run of dependencies whose output is already available from store',
)
@click.option(
    '--force-dependent/--no-force-dependent',
    help='Force run of all nodes depending on selected nodes, directly or indirectly',
)
@click.option('--keep/--no-keep', help='Keep uncommitted streams for debugging')
@click.option(
    '--keep-going/--no-keep-going',
    help='In case of an exception keep going with remaining datasets',
)
@click.option(
    '--detail/--no-detail', 'update_detail',
    help='Update detail view from stored node output',
)
@click.option(
    '--listing/--no-listing', 'update_listing',
    help='Update listing view from stored node output',
)
@click.option(
    '--cachesize', default=50, show_default=True,
    help='Number of messages to keep in memory for each stream',
)
@click.option(
    '--col', '--collection', 'collections', multiple=True,
    help='Run nodes for all datasets of selected collections, use "*" for all',
)
@click.argument('datasets', nargs=-1)
@click.pass_context
@click_async
async def marvcli_run(  # noqa: C901
        ctx, datasets, deps, excluded_nodes, force, force_dependent, force_deps, keep, keep_going,
        list_nodes, list_dependent, selected_nodes, update_detail, update_listing, cachesize,
        collections,
):
    """Run nodes for selected datasets.

    Datasets are specified by a list of set ids, or --collection
    <name>, use --collection=* to run for all collections. --node in
    conjunction with --collection=* will pick those collections for
    which the selected nodes are configured.

    Set ids may be abbreviated to any uniquely identifying
    prefix.
    """
    # pylint: disable=too-many-arguments,too-many-locals,too-many-branches,too-many-statements

    if collections and datasets:
        ctx.fail('--collection and DATASETS are mutually exclusive')

    if list_dependent and not selected_nodes:
        ctx.fail('--list-dependent needs at least one selected --node')

    if force_dependent and not selected_nodes:
        ctx.fail('--force-dependent needs at least one selected --node')

    if not any([datasets, collections, list_nodes]):
        click.echo(ctx.get_help())
        ctx.exit(1)

    deps = 'force' if force_deps else deps
    force = force_deps or force

    async with create_site() as site:
        if '*' in collections:
            if selected_nodes:
                collections = [k for k, v in site.collections.items()
                               if set(v.nodes).issuperset(selected_nodes)]
                if not collections:
                    ctx.fail('No collections have all selected nodes')
            else:
                collections = None
        else:
            for col in collections:
                if col not in site.collections:
                    ctx.fail(f'Unknown collection: {col}')

        if list_nodes:
            for col in (collections or sorted(site.collections.keys())):
                click.echo(f'{col}:')
                for name in sorted(site.collections[col].nodes):
                    if name == 'dataset':
                        continue
                    click.echo(f'    {name}')
            return

        if list_dependent:
            for col in (collections or sorted(site.collections.keys())):
                click.echo(f'{col}:')
                dependent = {x for name in selected_nodes
                             for x in site.collections[col].nodes[name].dependent}
                for name in sorted(x.name for x in dependent):
                    click.echo(f'    {name}')
            return

        errors = []
        if datasets:
            setids = datasets
        else:
            setids = await site.db.get_datasets_for_collections(collections)

        for setid in setids:
            if PDB:
                await site.run(setid, selected_nodes, deps, force, keep,
                               force_dependent, update_detail, update_listing,
                               excluded_nodes, cachesize=cachesize)
            else:
                try:
                    await site.run(setid, selected_nodes, deps, force, keep,
                                   force_dependent, update_detail, update_listing,
                                   excluded_nodes, cachesize=cachesize)
                except UnknownNode as e:
                    ctx.fail(f'Collection {e.args[0]} has no node {e.args[1]}')
                except DoesNotExist:
                    click.echo(f'ERROR: unknown {setid!r}', err=True)
                    if not keep_going:
                        raise
                except RequestedMessageTooOld as e:
                    _req = e.args[0]._requestor.node.name  # pylint: disable=no-member,protected-access
                    _handle = e.args[0].handle.node.name  # pylint: disable=no-member
                    click.echo(f"""
    ERROR: {_req} pulled {_handle} message {e.args[1]} not being in memory anymore.
    See https://ternaris.com/marv-robotics/docs/patterns.html#reduce-separately
    """, err=True)
                    ctx.abort()
                except BaseException as e:  # pylint: disable=broad-except
                    errors.append(setid)
                    if isinstance(e, KeyboardInterrupt):
                        log.warning('KeyboardInterrupt: aborting')
                        raise

                    if isinstance(e, DirectoryAlreadyExists):
                        click.echo(f"""
    ERROR: Directory for node run already exists:
    {e.args[0]!r}
    In case no other node run is in progress, this is a bug which you are kindly
    asked to report, providing information regarding any previous, failed node runs.
    """, err=True)
                        if not keep_going:
                            ctx.abort()
                    else:
                        log.error('Exception occured for dataset %s:', setid, exc_info=True)
                        log.error('Error occured for dataset %s: %s', setid, e)
                        if not keep_going:
                            ctx.exit(1)
        if errors:
            log.error('There were errors for %r', errors)


@marvcli.command('scan')
@click.option('-n', '--dry-run', is_flag=True)
@click_async
async def marvcli_scan(dry_run):
    """Scan for new and changed files."""
    async with create_site() as site:
        await site.scan(dry_run)


@marvcli.command('tag')
@click.option('--add', multiple=True, help='Tags to add')
@click.option('--rm', '--remove', 'remove', multiple=True, help='Tags to remove')
@click.argument('datasets', nargs=-1)
@click.pass_context
@click_async
async def marvcli_tag(ctx, add, remove, datasets):
    """Add or remove tags to datasets."""
    if not any([add, remove]) or not datasets:
        click.echo(ctx.get_help())
        ctx.exit(1)

    async with create_site() as site:
        await site.db.update_tags_by_setids(datasets, add, remove)


@marvcli.group('comment')
def marvcli_comment():
    """Add or remove comments."""


@marvcli_comment.command('add')
@click.option('-u', '--user', prompt=True, help='Commenting user')
@click.option('-m', '--message', required=True, help='Message for the comment')
@click.argument('datasets', nargs=-1)
@click_async
async def marvcli_comment_add(user, message, datasets):
    """Add comment as user for one or more datasets."""
    async with create_site() as site:
        try:
            await site.db.comment_by_setids(datasets, user, message)
        except DBError as e:
            err(f'ERROR: {e}', exit=1)


@marvcli_comment.command('list')
@click.argument('datasets', nargs=-1)
@click_async
async def marvcli_comment_list(datasets):
    """List comments for datasets.

    Output: setid comment_id date time author message
    """
    async with create_site() as site:
        comments = await site.db.get_comments_by_setids(datasets)
        for comment in sorted(comments, key=lambda x: (x.dataset[0].setid, x.id)):
            echo(comment.dataset[0].setid, comment.id,
                 datetime.datetime.fromtimestamp(int(comment.time_added / 1000)),  # noqa: DTZ
                 comment.author, repr(comment.text))


SHOW_TEMPLATE = Template("""
{% for dataset in datasets %}
- name: {{ dataset.name }}
  collection: {{ dataset.collection }}
  setid: {{ dataset.setid }}
  files:
  {% for file in dataset.files %}
    - path: {{ file.path }}
      size: {{ file.size }}
  {% endfor %}

{% endfor %}
""".strip(), trim_blocks=True, lstrip_blocks=True)


@marvcli.command('show')
@click.argument('datasets', nargs=-1, required=True)
@click_async
async def marvcli_show(datasets):
    """Show information for one or more datasets in form of a yaml document.

    Set ids may be abbreviated as long as they are unique.

    \b
    Some examples
      marv show setid  # show one dataset
      marv query --col=* | xargs marv show   # show all datasets
    """
    async with create_site() as site:
        datasets = await site.db.get_datasets_by_setids(datasets, prefetch=['files'])
        yamldoc = SHOW_TEMPLATE.render(datasets=datasets)
        echo(yamldoc, end='')


@marvcli_comment.command('rm')
@click.argument('ids', nargs=-1)
@click_async
async def marvcli_comment_rm(ids):
    """Remove comments.

    Remove comments by id as given in second column of: marv comment list
    """
    async with create_site() as site:
        await site.db.delete_comments_by_ids(ids)


@marvcli.group('user')
def marvcli_user():
    """Manage user accounts."""


@marvcli_user.command('add')
@click.option('--password', help='Password will be prompted')
@click.argument('username')
@click_async
async def marvcli_user_add(username, password):
    """Add a user."""
    if not USERGROUP_REGEX.match(username):
        click.echo(f'Invalid username: {username}', err=True)
        click.echo('Must only contain ASCII letters, numbers, and [-_+@.]', err=True)
        sys.exit(1)

    if password is None:
        password = click.prompt('Password', hide_input=True, confirmation_prompt=True)

    async with create_site() as site:
        try:
            await site.db.user_add(username, password, 'marv', '')
        except ValueError as e:
            click.echo(f'Error: {e.args[0]}', err=True)
            sys.exit(1)


@marvcli_user.command('list')
@click_async
async def marvcli_user_list():
    """List existing users."""
    async with create_site() as site:
        query = await site.db.get_users(deep=True)
        users = [(user.name, ', '.join(sorted(x.name for x in user.groups))) for user in query]
        users = [('User', 'Groups')] + users
        uwidth, _ = reduce(
            lambda xm_ym, x_y: (max(x_y[0], xm_ym[0]), max(x_y[1], xm_ym[1])),
            ((len(x), len(y)) for x, y in users),
        )
        fmt = '{:%ds} | {}' % uwidth
        for user, groups in users:
            click.echo(fmt.format(user, groups))


@marvcli_user.command('pw')
@click.option(
    '--password', prompt=True, hide_input=True, confirmation_prompt=True,
    help='Password will be prompted',
)
@click.argument('username')
@click.pass_context
@click_async
async def marvcli_user_pw(ctx, username, password):
    """Change password."""
    async with create_site() as site:
        try:
            await site.db.user_pw(username, password)
        except ValueError as e:
            ctx.fail(e.args[0])


@marvcli_user.command('rm')
@click.argument('username')
@click.pass_context
@click_async
async def marvcli_user_rm(ctx, username):
    """Remove a user."""
    async with create_site() as site:
        try:
            await site.db.user_rm(username)
        except ValueError as e:
            ctx.fail(e.args[0])


@marvcli.group('group')
def marvcli_group():
    """Manage user groups."""


@marvcli_group.command('add')
@click.argument('groupname')
@click.pass_context
@click_async
async def marvcli_group_add(ctx, groupname):
    """Add a group."""
    if not USERGROUP_REGEX.match(groupname):
        click.echo(f'Invalid groupname: {groupname}', err=True)
        click.echo('Must only contain ASCII letters, numbers, dash, underscore and dot', err=True)
        sys.exit(1)

    async with create_site() as site:
        try:
            await site.db.group_add(groupname)
        except ValueError as e:
            ctx.fail(e.args[0])


@marvcli_group.command('list')
@click_async
async def marvcli_group_list():
    """List existing groups."""
    async with create_site() as site:
        for group in await site.db.get_groups():
            click.echo(group.name)


@marvcli_group.command('adduser')
@click.argument('username')
@click.argument('groupname')
@click.pass_context
@click_async
async def marvcli_group_adduser(ctx, groupname, username):
    """Add an user to a group."""
    async with create_site() as site:
        try:
            await site.db.group_adduser(groupname, username)
        except ValueError as e:
            ctx.fail(e.args[0])


@marvcli_group.command('rmuser')
@click.argument('username')
@click.argument('groupname')
@click.pass_context
@click_async
async def marvcli_group_rmuser(ctx, groupname, username):
    """Remove an user from a group."""
    async with create_site() as site:
        try:
            await site.db.group_rmuser(groupname, username)
        except ValueError as e:
            ctx.fail(e.args[0])


@marvcli_group.command('rm')
@click.argument('groupname')
@click.pass_context
@click_async
async def marvcli_group_rm(ctx, groupname):
    """Remove a group."""
    async with create_site() as site:
        try:
            await site.db.group_rm(groupname)
        except ValueError as e:
            ctx.fail(e.args[0])


@marvcli.group('pip', hidden=not marv_ee)
def marvcli_pip():
    """Integrate pip commands (EE)."""


def ensure_python(siteconf, sitepackages):
    assert within_pyinstaller_bundle()
    pyexe = Path(code.__file__).parent / 'python'
    if not pyexe.exists():
        with pyexe.open('w') as f:
            f.write(f'#!/bin/sh\n'
                    f'export PYTHONPATH="{sitepackages}:$PYTHONPATH"\n'
                    f'exec {sys.executable} --config {siteconf} python "$@"')
        pyexe.chmod(0o700)
    sys.executable = str(pyexe)


@marvcli_pip.command('install', context_settings={'ignore_unknown_options': True})
@click.argument('pipargs', nargs=-1, type=click.UNPROCESSED)
def marvcli_pip_install(pipargs):
    """Install python package (EE).

    Use -e like with plain pip to install in editable mode.
    """
    assert within_pyinstaller_bundle()
    siteconf = get_site_config()
    config = make_config(siteconf)
    sitepackages = config.marv.sitepackages
    load_sitepackages(sitepackages)
    ensure_python(siteconf, sitepackages)
    from pip._internal.main import main as pip_main  # pylint: disable=import-outside-toplevel
    sys.argv = [sys.executable, 'install', '--prefix', config.marv.venv, *pipargs]
    sys.exit(pip_main())


@marvcli_pip.command('uninstall', context_settings={'ignore_unknown_options': True})
@click.argument('pipargs', nargs=-1, type=click.UNPROCESSED)
def marvcli_pip_uninstall(pipargs):
    """Uninstall python package (EE)."""
    assert within_pyinstaller_bundle()
    siteconf = get_site_config()
    sitepackages = make_config(siteconf).marv.sitepackages
    load_sitepackages(sitepackages)
    ensure_python(siteconf, sitepackages)
    from pip._internal.main import main as pip_main  # pylint: disable=import-outside-toplevel
    sys.argv = [sys.executable, 'uninstall', *pipargs]
    sys.exit(pip_main())


@marvcli.command('python', hidden=not marv_ee, context_settings={'ignore_unknown_options': True})
@click.option('-u', '--unbuffered', is_flag=True, help='Python -u equivalent (ignored)')
@click.option('-E', 'ignore', is_flag=True, help='Python -E equivalent (ignored)')
@click.option('-c', '--command', help='Python -c equivalent')
@click.argument('args', nargs=-1, type=click.UNPROCESSED)
def marvcli_python(unbuffered, ignore, command, args):  # pylint: disable=unused-argument
    """Drop into interactive python (EE)."""
    assert within_pyinstaller_bundle()
    siteconf = get_site_config()
    sitepackages = make_config(siteconf).marv.sitepackages
    load_sitepackages(sitepackages)
    ensure_python(siteconf, sitepackages)
    if command:
        sys.argv = ['-c', *args]
        sys.path.append('.')
        exec(compile(command, 'python_wrapper', 'exec'))  # pylint: disable=exec-used
    else:
        code.interact(local=locals())



# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

import json
from json import dumps as jsondumps

import pendulum
from aiohttp import web

from marv.collection import Filter
from marv.db import DBPermissionError, UnknownOperator
from marv.model import STATUS
from marv.utils import parse_datetime, parse_filesize, parse_timedelta
from .tooling import HTTPPermissionError
from .tooling import api_endpoint as marv_api_endpoint, get_global_granted, get_local_granted


ALIGN = {
    'filesize': 'right',
    'icon': 'center',
    'int': 'right',
    'timedelta': 'right',
}


FILTER_PARSER = {
    'datetime': lambda x: int(
        (parse_datetime(x) - parse_datetime('1970-01-01T00:00:00+00:00')).total_seconds() * 1000,
    ),
    'filesize': parse_filesize,
    'float': float,
    'int': int,
    'string': str,
    'string[]': str,
    'subset': lambda x: x,
    'timedelta': parse_timedelta,
    'words': lambda x: x.split(),
}


VALUE_TYPE_MAP = {
    'string[]': 'string',
    'words': 'string',
}


TIMEZONE = pendulum.local_timezone().name


# Order corresponds to marv.model.STATUS OrderedDict
STATUS_ICON = ['fire', 'eye-close', 'warning-sign', 'time']
STATUS_JSON = [jsondumps({'icon': STATUS_ICON[i], 'title': x},
                         separators=(',', ':'))
               for i, x in enumerate(STATUS.values())]
# TODO: reconsider in case we get a couple of more states
STATUS_STRS = {
    bitmask: ','.join(filter(None,
                             (STATUS_JSON[i] if bitmask & 2**i else None
                              for i in range(len(STATUS_JSON)))))
    for bitmask in range(2**(len(STATUS_JSON)))
}


def parse_filters(specs, filters):
    return [
        Filter(k,
               FILTER_PARSER[specs[k].value_type](v['val']),
               v['op'],
               specs[k].value_type)
        for k, v in filters.items()
    ]


@marv_api_endpoint('/meta', force_acl=['__unauthenticated__', '__authenticated__'], acl_key='list')
async def meta(request):
    site = request.app['site']
    if not set(request['user_groups']).intersection(request.app['route_acl']['list']):
        collections = []
    else:
        collections = await site.db.get_collections(user=request['username'])

    resp = web.json_response({
        'acl': get_global_granted(request),
        'collections': collections,
        'realms': list(site.config.marv.oauth),
        'timezone': TIMEZONE,
    })
    resp.headers['Cache-Control'] = 'no-cache'
    return resp


@marv_api_endpoint('/collection{_:/?}{collection_id:((?<=/).*)?}', acl_key='read')  # noqa: C901
async def collection(request):  # pylint: disable=too-many-locals  # noqa: C901
    site = request.app['site']
    collection_id = request.match_info['collection_id'] or site.collections.default_id

    try:
        all_known = await site.db.get_all_known_for_collection(site.collections, collection_id,
                                                               request['username'])
    except DBPermissionError:
        raise HTTPPermissionError(request)

    # try .. except for legacy reasons. will disappear with better data structuring
    try:
        collection = site.collections[collection_id]  # pylint: disable=redefined-outer-name
    except KeyError:
        raise web.HTTPNotFound()

    try:
        filters_dct = json.loads(request.query.get('filter', '{}'))
        filters = parse_filters(collection.filter_specs, filters_dct)
    except (KeyError, ValueError):
        raise web.HTTPBadRequest()

    try:
        rows = await site.db.get_filtered_listing(collection.table_descriptors, filters,
                                                  user=request['username'])
    except (KeyError, ValueError, UnknownOperator):
        raise web.HTTPBadRequest()

    real_id = next(
        x['id']
        for x in await site.db.get_collections(user=request['username'])
        if x['name'] == collection_id
    )
    acl = await site.db.get_acl('collection', real_id, request['username'],
                                get_local_granted(request))

    filters = [
        {
            'key': x.name,
            'constraints': all_known.get(x.name),
            'title': x.title,
            'operators': x.operators,
            'value_type': VALUE_TYPE_MAP.get(x.value_type, x.value_type),
        }
        for x in collection.filter_specs.values()
    ]

    def fmt(row):
        if not row['tag_value']:
            return '[]'
        return json.dumps(sorted(row['tag_value'].split(',')))

    rowdata = ',\n'.join([row['row'].replace('["#TAGS#"]', fmt(row))
                          .replace('"#TAGS#"', fmt(row)[1:-1] if fmt(row) != '[]' else '')
                          .replace('[,', '[')
                          .replace('"#STATUS#"', STATUS_STRS[row['status']])
                          for row in rows])
    dct = {
        'acl': acl,
        'all_known': all_known,
        'compare': bool(collection.compare),
        'filters': {'title': 'Filter',
                    'widget': {'type': 'filter', 'filters': filters}},
        'summary_config': {'title': 'Summary', 'items': collection.summary_items},
        'listing': {'title': f'Data sets ({len(rows)} found)',
                    'widget': {
                        'data': {
                            'columns': [{
                                'align': ALIGN.get(x.formatter, 'left'),
                                'formatter': x.formatter,
                                'sortkey': 'title' if x.formatter == 'route' else None,
                                'id': x.name,
                                'title': x.heading,
                                'list': x.islist,
                            } for x in collection.listing_columns],
                            'rows': ['#ROWS#'],
                            'sortcolumn': collection.sortcolumn,
                            'sortorder': collection.sortorder},
                        'type': 'table',
                    }},
    }
    indent = None
    separators = (',', ':')
    if request.app['debug']:
        indent = 2
        separators = (', ', ': ')
    jsondata = jsondumps(dct, indent=indent, separators=separators, sort_keys=True)
    jsondata = jsondata.replace('"#ROWS#"', rowdata)
    return web.Response(text=jsondata, headers={
        'Content-Type': 'application/json',
        'Cache-Control': 'no-cache',
    })



# Copyright 2016 - 2020  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

# pylint: disable=too-many-lines

import asyncio
import hashlib
import json
import re
import secrets
import string
import sys
from collections import namedtuple
from contextlib import asynccontextmanager
from contextvars import ContextVar
from datetime import datetime
from functools import reduce
from itertools import groupby, product
from logging import getLogger

import bcrypt
from pypika import JoinType, Order, SQLLiteQuery as Query, Table, Tables, Tuple, functions as fn
from pypika.terms import AggregateFunction, Criterion, EmptyCriterion, ValueWrapper
from tortoise import Tortoise as _Tortoise
from tortoise.exceptions import DoesNotExist, IntegrityError
from tortoise.transactions import current_transaction_map

from marv_node.setid import SetID
from . import utils
from .model import Group, Leaf, User
from .model import STATUS, STATUS_MISSING, STATUS_OUTDATED
from .model import __models__ as MODELS
from .utils import findfirst


log = getLogger(__name__)


USERGROUP_REGEX = re.compile(r'[0-9a-zA-Z\-_\.@+]+$')


class NoSetidFound(Exception):
    pass


class MultipleSetidFound(Exception):
    pass


class UnknownOperator(Exception):
    pass


class FilterError(Exception):
    pass


class DBError(Exception):
    pass


class DBPermissionError(Exception):
    pass


class DBConflictError(Exception):
    pass


class DBVersionError(Exception):
    pass


class DBNotInitialized(Exception):
    pass


@asynccontextmanager
async def scoped_session(database, txn=None):
    """Transaction scope for database operations."""
    if txn:
        yield txn
    else:
        connection = await database.connection_queue.get()
        try:
            # pylint: disable=protected-access
            async with connection._in_transaction() as txn:
                async def exq(query, count=False):
                    cnt, res = await txn.execute_query(query.get_sql())
                    return cnt if count else res
                txn.exq = exq
                yield txn
        finally:
            await database.connection_queue.put(connection)


class JsonListAgg:
    def __init__(self):
        self.items = set()

    def step(self, item):
        if item is not None:
            self.items.add(item)

    def finalize(self):
        return json.dumps(sorted(self.items))


async def create_or_ignore(tablename, txn, **kw):
    table = Table(tablename)
    query = Query.from_(table).select(table.star)
    for col, value in kw.items():
        query = query.where(getattr(table, col) == value)
    rows = await txn.exq(query)
    if not rows:
        await txn.exq(Query.into(table).columns(*kw.keys()).insert(*kw.values()))


def run_in_transaction(func):
    async def wrapper(database, *args, txn=None, **kwargs):
        async with scoped_session(database, txn=txn) as txn:
            return await func(database, *args, txn=txn, **kwargs)

    return wrapper


class FromExceptQuery:
    class Builder:
        def __init__(self, from_=None, except_=None):
            self._from = from_
            self._except = except_

        def from_(self, term):
            self._from = term
            return self

        def except_(self, term):
            self._except = term
            return self

        def get_sql(self, **kw):
            kw['subquery'] = False
            return f'({self._from.get_sql(**kw)} EXCEPT {self._except.get_sql(**kw)})'

    @classmethod
    def from_(cls, term):
        return cls.Builder(from_=term)

    @classmethod
    def except_(cls, term):
        return cls.Builder(except_=term)


class ValuesTuple(Tuple):
    def get_sql(self, **kw):
        return f'(VALUES {",".join(x.get_sql(**kw) for x in self.values)})'


class GroupConcat(AggregateFunction):
    def __init__(self, term, alias=None):
        super().__init__('GROUP_CONCAT', term, alias=alias)


class JsonGroupArray(AggregateFunction):
    def __init__(self, term, alias=None):
        super().__init__('JSON_GROUP_ARRAY', term, alias=alias)


class EscapableLikeCriterion(Criterion):
    operator = 'LIKE'

    def __init__(self, left, right, escchr, alias=None):
        super().__init__(alias)
        self.left = left
        self.right = right
        self.escchr = escchr

    def fields(self):
        return self.left.fields() + self.right.fields()

    def get_sql(self, with_alias=False, **kwargs):  # pylint: disable=arguments-differ
        sql = (f'{self.left.get_sql(**kwargs)} {self.operator} {self.right.get_sql(**kwargs)} '
               f'ESCAPE "{self.escchr}"')
        if with_alias and self.alias:
            return f'{sql} "{self.alias}"'
        return sql


class EscapableNotlikeCriterion(EscapableLikeCriterion):
    operator = 'NOT LIKE'


def esc(value, escchr='$'):
    return value.replace(f'{escchr}', f'{escchr}{escchr}')\
                .replace('_', f'{escchr}_')\
                .replace('%', f'{escchr}%')


def escaped_contains(field, value, escchr='$'):
    return EscapableLikeCriterion(field, ValueWrapper(f'%{esc(value, escchr)}%'), escchr)


def escaped_startswith(field, value, escchr='$'):
    return EscapableLikeCriterion(field, ValueWrapper(f'{esc(value, escchr)}%'), escchr)


def escaped_endswith(field, value, escchr='$'):
    return EscapableLikeCriterion(field, ValueWrapper(f'%{esc(value, escchr)}'), escchr)


def escaped_not_startswith(field, value, escchr='$'):
    return EscapableNotlikeCriterion(field, ValueWrapper(f'{esc(value, escchr)}%'), escchr)


OPS = {
    'between': lambda f, v: f.between(*v),
    'endswith': escaped_endswith,
    'eq': lambda f, v: f.eq(v),
    'gt': lambda f, v: f.gt(v),
    'gte': lambda f, v: f.gte(v),
    'ilike': lambda f, v: f.ilike(v),
    'in': lambda f, v: f.isin(v),
    'is': lambda f, v: f.isnull() if v is None else f.eq(v),
    'isnot': lambda f, v: f.notnull() if v is None else f.ne(v),
    'like': lambda f, v: f.like(v),
    'lt': lambda f, v: f.lt(v),
    'lte': lambda f, v: f.lte(v),
    'ne': lambda f, v: f.ne(v),
    'notbetween': lambda f, v: f.between(*v).negate(),
    'notilike': lambda f, v: f.not_ilike(v),
    'notin': lambda f, v: f.notin(v),
    'notlike': lambda f, v: f.not_like(v),
    'startswith': escaped_startswith,
    'substring': escaped_contains,
}


def resolve_filter(table, fltr, models):  # noqa: C901
    if not isinstance(fltr, dict):
        raise FilterError(f'Expected dict not {fltr!r}')

    operator = fltr.get('op')
    if not operator:
        raise FilterError(f"Missing 'op' in {fltr!r}")
    if not isinstance(operator, str):
        raise FilterError(f'Expected string not {operator!r}')

    value = fltr.get('value')
    name = fltr.get('name')
    tablemeta = findfirst(lambda x: x._meta.table == table._table_name, models)._meta

    if operator == 'and':
        return reduce(lambda x, y: x & resolve_filter(table, y, models), value, EmptyCriterion())
    if operator == 'or':
        return reduce(lambda x, y: x | resolve_filter(table, y, models), value, EmptyCriterion())
    if operator == 'not':
        return resolve_filter(table, value, models).negate()

    if not isinstance(name, str):
        raise FilterError(f"Missing 'name' in {fltr!r}")

    if '.' in name:
        fieldname, subname = name.split('.', 1)
        if fieldname in tablemeta.backward_fk_fields:
            field = tablemeta.fields_map[fieldname]
            subtable = Table(field.model_class._meta.table)
            resolved = resolve_filter(subtable, {'op': operator, 'name': subname, 'value': value},
                                      models)
            return getattr(table, 'id').isin(Query.from_(subtable)
                                             .select(getattr(subtable, field.relation_field))
                                             .where(resolved))

        if fieldname in tablemeta.m2m_fields:
            field = tablemeta.fields_map[fieldname]
            through = Table(field.through)
            rel = Table(field.model_class._meta.table)
            resolved = resolve_filter(rel, {'op': operator, 'name': subname, 'value': value},
                                      models)
            return getattr(table, 'id').isin(Query.from_(through)
                                             .join(rel)
                                             .on(getattr(through, field.forward_key) == rel.id)
                                             .where(resolved)
                                             .select(getattr(through, field.backward_key))
                                             .distinct())

        raise FilterError(f'Field {fieldname!r} not on model {tablemeta.table}')

    if name not in tablemeta.db_fields:
        raise FilterError(f'Field {name!r} not on model {tablemeta.table}')

    if operator not in OPS:
        raise FilterError(f'Unknown operator {operator!r}')

    field = getattr(table, name)
    return OPS[operator](field, value)


def cleanup_attrs(items):
    return [
        {k: x[k] for k in x.keys() if k not in ['acn_id', 'password']}
        for x in items
    ]


def cast_fields(keys, values):
    tmap = {
        'setid': SetID,
    }
    return [tmap.get(k, lambda x: x)(v) for k, v in zip(keys, values)]


def modelize(items, relations):
    res = []
    if not items:
        return res

    keys = items[0].keys()
    relidx = [i for i, x in enumerate(keys) if x == 'id'] + [len(keys)]

    dataset = namedtuple('dataset', keys[:relidx[1]] + list(relations))
    types = [
        namedtuple(x, keys[relidx[i+1]:relidx[i+2]])
        for i, x in enumerate(relations)
    ]

    for _, group in groupby(items, lambda x: x[0]):
        group = list(group)
        related = [
            sorted(set(filter(lambda x: x.id, (
                x(*y[relidx[i+1]:relidx[i+2]]) for y in group
            ))), key=lambda x: x.id)
            for i, x in enumerate(types)
        ]
        res.append(dataset(*(cast_fields(keys[:relidx[1]], group[0]) + related)))
    return res


def dt_to_sec(dtime):
    """Return seconds since epoch for datetime object."""
    sec = (datetime.fromisoformat(dtime) - datetime.utcfromtimestamp(0)).total_seconds()
    return int(sec)


async def get_items_for_query(query, txn):
    return [
        {k: x[k] for k in x.keys()}
        for x in (await txn.execute_query(query.get_sql()))[1]
    ]


async def process_items(query, delkeys, getkey, txn):
    items = await get_items_for_query(query, txn)
    res = {}
    for did, grp in groupby(items, key=lambda x: x['dataset_id']):
        assert did not in res
        res[did] = lst = []
        for item in grp:
            del item['id']
            del item['dataset_id']
            for delkey in delkeys:
                del item[delkey]
            lst.append(item[getkey] if getkey else item)
    return res


async def dump_users_groups(tables, dump, txn):
    group_t = tables.pop('group')
    user_t = tables.pop('user')
    user_group_t = tables.pop('user_group')
    groups = {}
    items = await get_items_for_query(Query.from_(user_group_t)
                                      .join(group_t)
                                      .on(user_group_t.group_id == group_t.id)
                                      .select('*')
                                      .where(~escaped_startswith(group_t.name, 'marv:'))
                                      .orderby(user_group_t.user_id)
                                      .orderby(group_t.name),
                                      txn)
    for uid, grp in groupby(items, key=lambda x: x['user_id']):
        assert uid not in groups
        groups[uid] = lst = []
        for group in grp:
            del group['id']
            del group['group_id']
            del group['user_id']
            name = group.pop('name')
            assert not group
            lst.append(name)

    dump['users'] = users = []
    items = await get_items_for_query(Query.from_(user_t)
                                      .select('*')
                                      .where(~escaped_startswith(user_t.name, 'marv:'))
                                      .orderby(user_t.name),
                                      txn)
    for user in items:
        uid = user.pop('id')
        user['groups'] = groups.pop(uid, [])
        user['time_created'] = dt_to_sec(user['time_created'])
        user['time_updated'] = dt_to_sec(user['time_updated'])
        users.append(user)
    assert not groups, groups


async def dump_leafs(tables, dump, txn):
    dump['leafs'] = leafs = []
    if 'leaf' in tables:
        leaf_t = tables.pop('leaf')
        items = await get_items_for_query(Query.from_(leaf_t)
                                          .select('*')
                                          .orderby(leaf_t.name),
                                          txn)
        for leaf in items:
            del leaf['id']
            leaf['time_created'] = dt_to_sec(leaf['time_created'])
            leaf['time_updated'] = dt_to_sec(leaf['time_updated'])
            leafs.append(leaf)


async def dump_acns(tables, dump, txn):  # pylint: disable=unused-argument
    del tables['acn']


async def dump_comments(tables, dump, txn):
    table = tables.pop('comment')
    dump['comments'] = await process_items(Query.from_(table)
                                           .select(table.star)
                                           .orderby(table.dataset_id)
                                           .orderby(table.id),
                                           (),
                                           None,
                                           txn)


async def dump_files(tables, dump, txn):
    table = tables.pop('file')
    dump['files'] = await process_items(Query.from_(table)
                                        .select(table.star)
                                        .orderby(table.dataset_id)
                                        .orderby(table.id),
                                        ('idx',),
                                        None,
                                        txn)


async def dump_tags(tables, dump, txn):
    dataset_tag_t = tables.pop('dataset_tag')
    tag_t = tables.pop('tag')
    dump['tags'] = await process_items(Query.from_(dataset_tag_t)
                                       .join(tag_t)
                                       .on(dataset_tag_t.tag_id == tag_t.id)
                                       .select('*')
                                       .orderby(dataset_tag_t.dataset_id)
                                       .orderby(tag_t.value),
                                       ('tag_id',),
                                       'value',
                                       txn)


async def dump_datasets(tables, dump, txn):
    comments = dump.pop('comments')
    files = dump.pop('files')
    tags = dump.pop('tags')

    dump['datasets'] = collections = {}
    collection_t = tables.pop('collection')
    dataset_t = tables.pop('dataset')
    items = await get_items_for_query(Query.from_(dataset_t)
                                      .join(collection_t)
                                      .on(dataset_t.collection_id == collection_t.id)
                                      .select(dataset_t.star, collection_t.name.as_('collection'))
                                      .orderby(dataset_t.setid),
                                      txn)
    for dataset in items:
        did = dataset.pop('id')
        del dataset['acn_id']
        del dataset['collection_id']
        dataset['comments'] = comments.pop(did, [])
        dataset['files'] = files.pop(did)
        dataset['tags'] = tags.pop(did, [])
        collections.setdefault(dataset.pop('collection'), []).append(dataset)

    assert not comments, comments
    assert not files, files
    assert not tags, tags


async def dump_metadata(tables, dump, txn):  # pylint: disable=unused-argument
    tables.pop('metadata')


async def restore_users(site, dct, txn):
    users = dct.pop('users')
    for user in users or []:
        user.setdefault('realm', 'marv')
        user.setdefault('realmuid', '')
        groups = user.pop('groups', [])
        await site.db.user_add(_restore=True, txn=txn, **user)
        for grp in groups:
            try:
                await site.db.group_adduser(grp, user['name'], txn=txn)
            except ValueError:
                await site.db.group_add(grp, txn=txn)
                await site.db.group_adduser(grp, user['name'], txn=txn)


async def restore_leafs(site, dct, txn):
    leafs = dct.pop('leafs')
    for leaf in leafs or []:
        await site.db.leaf_add(_restore=True, txn=txn, **leaf)


async def restore_datasets(site, dct, txn):
    datasets = dct.pop('datasets')
    for key, sets in datasets.items():
        await site.collections[key].restore_datasets(sets, txn)


class Tortoise(_Tortoise):
    @classmethod
    def _discover_models(cls, model, app_label):  # pylint: disable=arguments-differ
        model._meta.app = app_label
        model._meta.finalise_pk()
        return [model]


class Database:
    # pylint: disable=too-many-public-methods

    VERSION = '20.06'

    EXPORT_HANDLERS = (
        ({'group', 'user', 'user_group'}, dump_users_groups),
        ({'leaf'}, dump_leafs),
        ({'acn'}, dump_acns),
        ({'comment'}, dump_comments),
        ({'file'}, dump_files),
        ({'tag', 'dataset_tag'}, dump_tags),
        ({'dataset', 'collection'}, dump_datasets),
        ({'metadata'}, dump_metadata),
    )

    IMPORT_HANDLERS = (
        ({'users'}, restore_users),
        ({'leafs'}, restore_leafs),
        ({'datasets'}, restore_datasets),
    )

    MODELS = MODELS

    def __init__(self):
        self.connections = []
        self.connection_queue = asyncio.Queue()

    async def initialize_connections(self):
        defcon = Tortoise.get_connection('default')
        for idx in range(48):
            name = f'connection_{idx}'
            db_params = defcon.pragmas.copy()
            db_params['connection_name'] = name
            connection = defcon.__class__(defcon.filename, **db_params)
            await connection.create_connection(with_db=True)

            await connection._connection._execute(  # pylint: disable=protected-access
                connection._connection._conn.create_aggregate,  # pylint: disable=protected-access
                'JSON_GROUP_ARRAY', 1, JsonListAgg,
            )

            current_transaction_map[name] = ContextVar(name, default=connection)
            await self.connection_queue.put(connection)
            self.connections.append(connection)

    async def close_connections(self):
        for connection in self.connections:
            # pylint: disable=protected-access
            cursor = await connection._connection.execute('PRAGMA optimize')
            await cursor.close()
            await connection.close()
            current_transaction_map.pop(connection.connection_name)
        self.connections.clear()

    @run_in_transaction
    async def authenticate(self, username, password, txn=None):
        if not username or not password:
            return False
        try:
            user = await User.get(name=username, realm='marv').using_db(txn)
        except DoesNotExist:
            return False
        return bcrypt.checkpw(password.encode('utf-8'), user.password.encode('utf-8'))

    @run_in_transaction   # noqa: C901
    async def bulk_um(self, users_add, users_remove, groups_add, groups_remove,    # noqa: C901
                      groups_add_users, groups_remove_users, txn=None):
        # pylint: disable=too-many-arguments
        try:
            everybody = await Group.get(name='marv:users').using_db(txn)

            for name in users_remove:
                user = await User.get(name=name).using_db(txn)
                group = await Group.get(name=f'marv:user:{name}').using_db(txn)
                await user.groups.remove(group, everybody, using_db=txn)
                await user.delete(using_db=txn)
                await group.delete(using_db=txn)

            for name in users_add:
                if not USERGROUP_REGEX.match(name):
                    raise DBError('User name can only contain alphanumeric characters and [-_+@.]')
                user = await User.create(name=name, realm='marv', using_db=txn)
                await user.groups.add(
                    await Group.create(name=f'marv:user:{name}', using_db=txn),
                    everybody,
                    using_db=txn,
                )

            for name in groups_remove:
                await Group.get(name=name).using_db(txn).delete()

            for name in groups_add:
                if not USERGROUP_REGEX.match(name):
                    raise DBError(
                        'Group name can only contain alphanumeric characters and [-_+@.]',
                    )
                await Group.create(name=name, using_db=txn)

            for name, users in groups_remove_users:
                group = await Group.get(name=name).using_db(txn)
                users = await User.filter(name__in=users).using_db(txn)
                await group.users.remove(*users, using_db=txn)

            for name, users in groups_add_users:
                group = await Group.get(name=name).using_db(txn)
                users = await User.filter(name__in=users).using_db(txn)
                await group.users.add(*users, using_db=txn)
        except DoesNotExist:
            raise DBError('Entity does not exist')

    @run_in_transaction
    async def user_add(self, name, password, realm, realmuid, given_name=None, family_name=None,
                       email=None, time_created=None, time_updated=None, _restore=None, txn=None):
        # pylint: disable=too-many-arguments
        if not USERGROUP_REGEX.match(name):
            raise DBError('User name can only contain alphanumeric characters and [-_+@.]')
        if not _restore and password is not None:
            password = bcrypt.hashpw(password.encode('utf-8'), bcrypt.gensalt()).decode('utf-8')
        now = datetime.utcnow()  # noqa: DTZ
        time_created = datetime.utcfromtimestamp(time_created) if time_created else now
        time_updated = datetime.utcfromtimestamp(time_updated) if time_updated else now
        try:
            user = await User.create(name=name, password=password, realm=realm,
                                     given_name=given_name, family_name=family_name,
                                     email=email, realmuid=realmuid, time_created=time_created,
                                     time_updated=time_updated, using_db=txn)
            if name != 'marv:anonymous':
                everybody = await Group.get(name='marv:users').using_db(txn)
                await user.groups.add(
                    await Group.create(name=f'marv:user:{name}', using_db=txn),
                    everybody,
                    using_db=txn,
                )
            if _restore:
                user = Table('user')
                await txn.exq(Query.update(user)
                              .set(user.time_updated, time_updated)
                              .where(user.name == name))
        except IntegrityError:
            raise ValueError(f'User {name} exists already')
        return user

    @run_in_transaction
    async def user_rm(self, username, txn=None):
        try:
            user = await User.get(name=username).using_db(txn)
            group = await Group.get(name=f'marv:user:{username}').using_db(txn)
            everybody = await Group.get(name='marv:users').using_db(txn)
            await user.groups.remove(group, everybody, using_db=txn)
            await user.delete(using_db=txn)
            await group.delete(using_db=txn)

            await user.delete(using_db=txn)
        except DoesNotExist:
            raise ValueError(f'User {username} does not exist')

    @run_in_transaction
    async def user_pw(self, username, password, txn=None):
        try:
            user = await User.get(name=username).using_db(txn)
        except DoesNotExist:
            raise ValueError(f'User {username} does not exist')

        user.password = bcrypt.hashpw(password.encode('utf-8'), bcrypt.gensalt()).decode('utf-8')
        user.time_updated = int(utils.now())
        await user.save(using_db=txn)

    @run_in_transaction
    async def group_add(self, groupname, txn=None):
        if not USERGROUP_REGEX.match(groupname):
            raise DBError('Group name can only contain alphanumeric characters and [-_+@.]')
        try:
            await Group.create(name=groupname, using_db=txn)
        except IntegrityError:
            raise ValueError(f'Group {groupname} exists already')

    @run_in_transaction
    async def group_rm(self, groupname, txn=None):
        try:
            group = await Group.get(name=groupname).using_db(txn)
            await group.delete(using_db=txn)
        except DoesNotExist:
            raise ValueError(f'Group {groupname} does not exist')

    @run_in_transaction
    async def group_adduser(self, groupname, username, txn=None):
        try:
            group = await Group.get(name=groupname).using_db(txn)
        except DoesNotExist:
            raise ValueError(f'Group {groupname} does not exist')
        try:
            user = await User.get(name=username).using_db(txn)
        except DoesNotExist:
            raise ValueError(f'User {username} does not exist')
        await group.users.add(user, using_db=txn)

    @run_in_transaction
    async def group_rmuser(self, groupname, username, txn=None):
        try:
            group = await Group.get(name=groupname).using_db(txn)
        except DoesNotExist:
            raise ValueError(f'Group {groupname} does not exist')
        try:
            user = await User.get(name=username).using_db(txn)
        except DoesNotExist:
            raise ValueError(f'User {username} does not exist')
        await group.users.remove(user, using_db=txn)

    @run_in_transaction
    async def get_users(self, deep=False, txn=None):
        query = User.all().using_db(txn).order_by('name')
        if deep:
            query = query.prefetch_related('groups')
        return await query

    @run_in_transaction
    async def get_user_by_name(self, name, deep=False, txn=None):
        query = User.filter(name=name).using_db(txn)
        if deep:
            query = query.prefetch_related('groups')
        return await query.first()

    @run_in_transaction
    async def get_user_by_realmuid(self, realm, realmuid, deep=False, txn=None):
        query = User.filter(realm=realm, realmuid=realmuid).using_db(txn)
        if deep:
            query = query.prefetch_related('groups')
        return await query.first()

    @run_in_transaction
    async def get_groups(self, deep=False, txn=None):
        query = Group.all().using_db(txn).order_by('name')
        if deep:
            query = query.prefetch_related('users')
        return await query

    @run_in_transaction
    async def leaf_add(self, name, access_token=None, refresh_token=None, time_created=None,
                       time_updated=None, _restore=None, txn=None):
        # pylint: disable=too-many-arguments
        if _restore:
            clear_access_token = None
            clear_refresh_token = None
        else:
            alnum = string.ascii_letters + string.digits
            clear_access_token = ''.join(secrets.choice(alnum) for i in range(20))
            clear_refresh_token = ''.join(secrets.choice(alnum) for i in range(20))
            access_token = hashlib.sha256(clear_access_token.encode()).hexdigest()
            refresh_token = hashlib.sha256(clear_refresh_token.encode()).hexdigest()

        now = datetime.utcnow()  # noqa: DTZ
        time_created = datetime.utcfromtimestamp(time_created) if time_created else now
        time_updated = datetime.utcfromtimestamp(time_updated) if time_updated else now
        try:
            leaf = await Leaf.create(name=name, access_token=access_token,
                                     refresh_token=refresh_token, time_created=time_created,
                                     time_updated=time_updated, using_db=txn)
            if _restore:
                leaf_t = Table('leaf')
                await txn.exq(Query.update(leaf_t)
                              .set(leaf_t.time_updated, time_updated)
                              .where(leaf_t.name == name))
        except IntegrityError:
            raise ValueError(f'Recording unit {name} exists already')

        return leaf

    @run_in_transaction
    async def leaf_regentoken(self, name, txn=None):
        try:
            leaf = await Leaf.get(name=name).using_db(txn)
        except DoesNotExist:
            raise ValueError(f'Recording unit {name} does not exist')

        alnum = string.ascii_letters + string.digits
        clear_access_token = ''.join(secrets.choice(alnum) for i in range(20))
        clear_refresh_token = ''.join(secrets.choice(alnum) for i in range(20))
        leaf.access_token = hashlib.sha256(clear_access_token.encode('utf8')).hexdigest()
        leaf.refresh_token = hashlib.sha256(clear_refresh_token.encode('utf8')).hexdigest()
        leaf.time_updated = int(utils.now())
        await leaf.save(using_db=txn)
        leaf.clear_access_token = clear_access_token
        leaf.clear_refresh_token = clear_refresh_token
        return leaf

    @run_in_transaction
    async def get_leafs(self, txn=None):
        return await Leaf.all().using_db(txn).order_by('name')

    @run_in_transaction
    async def get_leaf_by_token(self, clear_access_token, txn=None):
        access_token = hashlib.sha256(clear_access_token.encode()).hexdigest()
        return await Leaf.filter(access_token=access_token).using_db(txn).first()

    @run_in_transaction
    async def get_acl(self, model, id, user, default, txn=None):
        # pylint: disable=unused-argument, too-many-arguments
        return default

    @staticmethod
    def get_actionable(modelname, user, action):
        # pylint: disable=unused-argument
        model = Table(modelname)
        return Query.from_(model).select(model.id)

    @staticmethod
    def get_resolve_value_query(rel, relations, user, action):
        # pylint: disable=unused-argument
        out, tmp = Tables('out', 'tmp')
        return Query.with_(Query.with_(Query.from_(ValuesTuple(*relations))
                                       .select('*'), 'tmp(value, back_id)')
                           .from_(tmp)
                           .join(rel)
                           .on(tmp.value == rel.value)
                           .select(rel.id, tmp.back_id), 'out(rel_id, back_id)')\
                    .from_(out)\
                    .select(out.rel_id, out.back_id)

    @staticmethod
    def id_crit(ids, user, action):  # pylint: disable=unused-argument
        dataset = Table('dataset')
        return dataset.id.isin(ids)

    @staticmethod
    def setid_crit(ids, user, action):  # pylint: disable=unused-argument
        dataset = Table('dataset')
        return dataset.setid.isin([str(x) for x in ids])

    @run_in_transaction
    async def resolve_shortids(self, prefixes, discarded=False, txn=None):
        setids = set()
        dataset = Table('dataset')
        for prefix in prefixes:
            if isinstance(prefix, SetID):
                prefix = str(prefix)

            setid = await txn.exq(Query.from_(dataset)
                                  .select('setid')
                                  .where(escaped_startswith(dataset.setid, prefix)
                                         & (dataset.discarded == discarded))
                                  .limit(2))

            if not setid:
                discarded = 'discarded ' if discarded else ''
                raise NoSetidFound(f'{prefix} does not match any {discarded}dataset')
            if len(setid) > 1:
                matches = '\n  '.join([f'{x}' for x in setid])
                raise MultipleSetidFound(f'{prefix} matches multiple:\n'
                                         f'  {matches}\n')
            setids.add(SetID(setid[0][0]))
        return sorted(setids)

    @run_in_transaction
    async def get_collections(self, user, txn=None):  # pylint: disable=unused-argument
        collection = Table('collection')
        return [
            {'id': x, 'name': y}
            for x, y in await txn.exq(Query.from_(collection)
                                      .select(collection.id, collection.name)
                                      .where(collection.id.isin(
                                          self.get_actionable('collection', user, 'list')))
                                      .orderby(collection.name))
        ]

    async def _get_datasets_by_crit(self, crit, prefetch, txn):
        dataset = Table('dataset')
        query = Query.from_(dataset).where(crit).select(dataset.star)
        for related in prefetch:
            table = Table(related[:-1])
            if related == 'collections':
                query = query.join(table)\
                             .on(dataset.collection_id == table.id)\
                             .select(table.star)
            elif related == 'tags':
                through = Table('dataset_tag')
                query = query.join(through, how=JoinType.left_outer)\
                             .on(dataset.id == through.dataset_id)\
                             .join(table, how=JoinType.left_outer)\
                             .on(through.tag_id == table.id)\
                             .select(table.star)
            else:
                query = query.join(table, how=JoinType.left_outer)\
                             .on(table.dataset_id == dataset.id)\
                             .select(table.star)
        items = await txn.exq(query)
        return modelize(items, prefetch)

    @run_in_transaction
    async def get_datasets_by_setids(self, setids, prefetch, user, action=None, txn=None):
        # pylint: disable=too-many-arguments
        ret = await self._get_datasets_by_crit(self.setid_crit(setids, user, action), prefetch, txn)
        if len(ret) != len(setids):
            raise DBPermissionError
        return ret

    @run_in_transaction
    async def get_datasets_by_dbids(self, ids, prefetch, user, action=None, txn=None):
        # pylint: disable=too-many-arguments
        ret = await self._get_datasets_by_crit(self.id_crit(ids, user, action), prefetch, txn)
        if len(ret) != len(ids):
            raise DBPermissionError
        return ret

    @run_in_transaction
    async def get_filepath_by_setid_idx(self, setid, idx, user, txn=None):
        dataset, file = Tables('dataset', 'file')
        res = await txn.exq(Query.from_(file)
                            .join(dataset)
                            .on(file.dataset_id == dataset.id)
                            .where((file.idx == idx)
                                   & self.setid_crit([setid], user, 'download_raw'))
                            .select('path'))
        if not res:
            raise DBPermissionError
        return res[0]['path']

    @run_in_transaction
    async def get_datasets_for_collections(self, collections, txn=None):
        dataset = Table('dataset')
        bitmask = ValueWrapper(STATUS_MISSING)
        query = Query.from_(dataset)\
                     .select('setid')\
                     .where((dataset.discarded.eq(False))
                            & dataset.status.bitwiseand(bitmask).ne(bitmask))
        if collections is not None:
            collection = Table('collection')
            query = query.join(collection)\
                         .on(dataset.collection_id == collection.id)\
                         .where(collection.name.isin(collections))
        return [SetID(x['setid']) for x in await txn.exq(query)]

    async def _set_dataset_discarded_by_crit(self, crit, state, txn):
        dataset = Table('dataset')
        return await txn.exq(Query.update(dataset)
                             .set(dataset.discarded, state)
                             .where(crit), count=True)

    @run_in_transaction
    async def discard_datasets_by_setids(self, setids, state=True, txn=None):
        await self._set_dataset_discarded_by_crit(self.setid_crit(setids, '::', None), state, txn)

    @run_in_transaction
    async def discard_datasets_by_dbids(self, ids, state, user, txn=None):
        count = await self._set_dataset_discarded_by_crit(self.id_crit(ids, user, 'delete'),
                                                          state, txn)
        if count != len(ids):
            raise DBPermissionError

    @run_in_transaction
    async def cleanup_discarded(self, descs, txn=None):
        collection, dataset = Tables('collection', 'dataset')
        datasets = await txn.exq(Query.from_(dataset)
                                 .join(collection)
                                 .on(dataset.collection_id == collection.id)
                                 .select(collection.name, dataset.id)
                                 .where(dataset.discarded.eq(True))
                                 .orderby(collection.name))
        if not datasets:
            return

        datasets = [
            (col, [x[1] for x in tuples])
            for col, tuples in groupby(datasets, lambda x: x[0])
        ]
        for col, ids in datasets:
            await txn.exq(Query.from_(dataset)
                          .where(dataset.id.isin(ids))
                          .delete())

            for table in (Table('dataset_tag'), Table('comment'), Table('file')):
                await txn.exq(Query.from_(table)
                              .where(table.dataset_id.isin(ids))
                              .delete())

            tbls = descs[col]
            assert not tbls[0].through, 'Listing table should be first'
            listing = Table(tbls[0].table)

            await txn.exq(Query.from_(listing)
                          .where(listing.id.isin(ids))
                          .delete())

            for desc in [x for x in tbls if x.through]:
                through = Table(desc.through)
                id_field = getattr(through, desc.listing_id)
                await txn.exq(Query.from_(through)
                              .where(id_field.isin(ids))
                              .delete())

    @run_in_transaction
    async def bulk_comment(self, comments, user, txn=None):
        comment, dataset, tmp, user_t = Tables('comment', 'dataset', 'tmp', 'user')
        comments = [(x['dataset_id'], x['author'], x['time_added'], x['text']) for x in comments]
        query = Query.into(comment)\
                     .columns('dataset_id', 'author', 'time_added', 'text')\
                     .from_(Query.with_(Query.from_(ValuesTuple(*comments))
                                        .select('*'), 'tmp(dataset_id, author, time_added, text)')
                            .from_(tmp)
                            .join(dataset)
                            .on(tmp.dataset_id == dataset.id)
                            .where(self.id_crit([x[0] for x in comments], user, 'comment'))
                            .join(user_t)
                            .on(tmp.author == user_t.name)
                            .select(tmp.star))\
                     .select('*')
        count = await txn.exq(query, count=True)
        if count != len(comments):
            raise DBPermissionError

    @run_in_transaction
    async def comment_by_setids(self, setids, author, text, txn=None):
        comment, dataset, tmp, user = Tables('comment', 'dataset', 'tmp', 'user')
        time_added = int(utils.now() * 1000)
        query = Query.into(comment)\
                     .columns('dataset_id', 'author', 'time_added', 'text')\
                     .from_(Query.with_(Query.from_(ValuesTuple((author, time_added, text)))
                                        .select('*'), 'tmp(author, time_added, text)')
                            .from_(dataset)
                            .join(tmp)
                            .cross()
                            .join(user)
                            .on(tmp.author == user.name)
                            .where(dataset.setid.isin([str(x) for x in setids]))
                            .select(dataset.id, tmp.star))\
                     .select('*')
        count = await txn.exq(query, count=True)
        if count != len(setids):
            raise DBError(f'Commenting failed. User {author!r} or one of the datasets are missing')

    @run_in_transaction
    async def delete_comments_by_ids(self, ids, txn=None):
        comment = Table('comment')
        await txn.exq(Query.from_(comment)
                      .where(comment.id.isin(ids))
                      .delete())

    @run_in_transaction
    async def get_comments_by_setids(self, setids, txn=None):
        comment, dataset = Tables('comment', 'dataset')
        if setids:
            crit = dataset.setid.isin([str(x) for x in setids])
        else:
            crit = dataset.discarded.ne(True)
        query = Query.from_(comment)\
                     .join(dataset).on(comment.dataset_id == dataset.id)\
                     .select(comment.star, dataset.star)\
                     .where(crit)
        items = await txn.exq(query)
        return modelize(items, ['dataset'])

    async def _ensure_values(self, tablename, values, user, txn):
        # pylint: disable=unused-argument
        table = Table(tablename)
        query = Query.into(table)\
                     .columns('value')\
                     .from_(FromExceptQuery
                            .from_(Query.from_(ValuesTuple(*values)).select('*'))
                            .except_(Query().from_(table).select(table.value)))\
                     .select('*')
        return await txn.exq(query, count=True)

    async def _ensure_refs(self, relname, throughname, rel_id, back_id, relations, user,
                           action, txn):
        # pylint: disable=too-many-arguments
        rel, through = Tables(relname, throughname)
        relid = getattr(through, rel_id)
        backid = getattr(through, back_id)
        query = Query.into(through)\
                     .columns(rel_id, back_id)\
                     .from_(FromExceptQuery
                            .from_(self.get_resolve_value_query(rel, relations, user, action))
                            .except_(Query.from_(through).select(relid, backid)))\
                     .select('*')
        return await txn.exq(query, count=True)

    async def _delete_values_without_ref(self, relname, throughname, rel_id, txn):
        rel, through = Tables(relname, throughname)
        await txn.exq(Query.from_(rel)
                      .where(rel.id.notin(Query.from_(throughname)
                                          .select(getattr(through, rel_id))
                                          .distinct()))
                      .delete())

    @run_in_transaction
    async def bulk_tag(self, add, remove, user, txn=None):
        if add:
            await self._ensure_values('tag', [(x[0],) for x in add], user, txn)
            count = await self._ensure_refs('tag', 'dataset_tag', 'tag_id', 'dataset_id', add,
                                            user, 'tag', txn)
            if count != len(add):
                raise DBPermissionError

        if remove:
            through, tag = Tables('dataset_tag', 'tag')
            query = Query.from_(through)\
                         .delete()\
                         .select(through.tag_id, through.dataset_id)\
                         .where(Tuple(through.tag_id, through.dataset_id)
                                .isin(self.get_resolve_value_query(tag, remove, user, 'tag')))
            count = await txn.exq(query, count=True)
            if count != len(remove):
                raise DBPermissionError

    @run_in_transaction
    async def update_tags_by_setids(self, setids, add, remove, txn=None):
        setids = [str(x) for x in setids]
        dataset = Table('dataset')
        colids = await txn.exq(Query.from_(dataset)
                               .where(dataset.setid.isin(setids))
                               .select(dataset.id))
        add = [(tag, colid['id']) for tag, colid in product(add, colids)]
        remove = [(tag, colid['id']) for tag, colid in product(remove, colids)]
        await self.bulk_tag(add, remove, user='::', txn=txn)

    @run_in_transaction
    async def list_tags(self, collections=None, txn=None):
        collection, dataset, dataset_tag, tag = Tables('collection', 'dataset', 'dataset_tag',
                                                       'tag')
        query = Query.from_(tag)\
                     .select('value')\
                     .distinct()\
                     .orderby('value')

        if collections:
            query = query.join(dataset_tag, how=JoinType.left_outer)\
                         .on(tag.id == dataset_tag.tag_id)\
                         .join(dataset)\
                         .on(dataset_tag.dataset_id == dataset.id)\
                         .join(collection)\
                         .on(dataset.collection_id == collection.id)\
                         .where(collection.name.isin(collections))

        return [x['value'] for x in await txn.exq(query)]

    @run_in_transaction
    async def delete_comments_tags(self, setids, comments, tags, txn=None):
        dataset, dataset_tag, comment = Tables('dataset', 'dataset_tag', 'comment')
        subq = Query.from_(dataset)\
                    .select(dataset.id)\
                    .where(dataset.setid.isin([str(x) for x in setids]))
        if tags:
            await txn.exq(Query.from_(dataset_tag)
                          .where(dataset_tag.dataset_id.isin(subq))
                          .delete())

        if comments:
            await txn.exq(Query.from_(comment)
                          .where(comment.dataset_id.isin(subq))
                          .delete())

    @run_in_transaction
    async def delete_tag_values_without_ref(self, txn=None):
        await self._delete_values_without_ref('tag', 'dataset_tag', 'tag_id', txn)

    @run_in_transaction
    async def get_all_known_for_collection(self, collections, name, user, txn=None):
        collection_t = Table('collection')
        res = await txn.exq(self.get_actionable('collection', user, 'read')
                            .where(collection_t.name == name))
        if len(res) != 1:
            raise DBPermissionError
        collection = collections[name]
        descs = [x for x in collection.table_descriptors if x.through]
        all_known = {
            desc.key: [
                x['value'] for x in await txn.exq(Query.from_(desc.table).select('value'))
            ]
            for desc in descs
            if {'any', 'all'}.intersection(collection.filter_specs[desc.key].operators)
        }
        all_known.update({
            'status': list(STATUS),
            'tags': await self.get_all_known_tags_for_collection(collection.name),
        })
        return all_known

    @run_in_transaction
    async def get_all_known_tags_for_collection(self, collection_name, txn=None):
        collection, dataset, dataset_tag, tag = Tables('collection', 'dataset', 'dataset_tag',
                                                       'tag')
        query = Query.from_(tag)\
                     .select(tag.value)\
                     .where(tag.id.isin(Query.from_(dataset_tag)
                                        .join(dataset)
                                        .on(dataset_tag.dataset_id == dataset.id)
                                        .select(dataset_tag.tag_id)
                                        .join(collection)
                                        .on(dataset.collection_id == collection.id)
                                        .where(collection.name == collection_name)
                                        .distinct()))\
                     .orderby(tag.value)
        return [x['value'] for x in await txn.exq(query)]

    @run_in_transaction  # noqa: C901
    async def get_filtered_listing(self, descs, filters, user, txn=None):  # noqa: C901
        # pylint: disable=too-many-locals,too-many-branches,too-many-statements,unused-argument
        listing, dataset, dataset_tag, tag = \
            Tables(descs[0].table, 'dataset', 'dataset_tag', 'tag')

        query = Query.from_(listing)\
                     .join(dataset, how=JoinType.left_outer)\
                     .on(listing.id == dataset.id)\
                     .join(dataset_tag, how=JoinType.left_outer)\
                     .on(dataset.id == dataset_tag.dataset_id)\
                     .join(tag, how=JoinType.left_outer)\
                     .on(dataset_tag.tag_id == tag.id)\
                     .select(listing.id.as_('id'),
                             listing.row.as_('row'),
                             dataset.status.as_('status'),
                             GroupConcat(tag.value).as_('tag_value'))\
                     .where(dataset.discarded.ne(True)
                            & dataset.id.isin(self.get_actionable('dataset', user, 'list')))

        for name, value, operator, val_type in filters:
            if isinstance(value, int):
                value = min(value, sys.maxsize)

            if name == 'comments':
                comment = Table('comment')
                query = query.where(listing.id.isin(Query.from_(comment)
                                                    .select(comment.dataset_id)
                                                    .where(escaped_contains(comment.text, value))))
                continue

            if name == 'status':
                status_ids = list(STATUS.keys())
                bitmasks = [2**status_ids.index(x) for x in value]
                bitmask = sum(bitmasks)
                if operator == 'any':
                    query = query.where(dataset.status.bitwiseand(bitmask))

                elif operator == 'all':
                    query = query.where(dataset.status.bitwiseand(bitmask) == bitmask)

                else:
                    raise UnknownOperator(operator)

                continue

            if name == 'tags':
                if operator == 'any':
                    query = query.where(listing.id.isin(Query.from_(dataset_tag)
                                                        .join(tag)
                                                        .on(dataset_tag.tag_id == tag.id)
                                                        .where(tag.value.isin(value))
                                                        .select(dataset_tag.dataset_id)
                                                        .distinct()))

                elif operator == 'all':
                    query = query.where(listing.id.isin(Query.from_(dataset_tag)
                                                        .join(tag)
                                                        .on(dataset_tag.tag_id == tag.id)
                                                        .where(tag.value.isin(value))
                                                        .select(dataset_tag.dataset_id)
                                                        .groupby(dataset_tag.dataset_id)
                                                        .having(fn.Count('*') == len(value))))

                else:
                    raise UnknownOperator(operator)

                continue

            if val_type == 'datetime':
                if operator == 'eq':
                    field = getattr(listing, name)
                    query = query.where(field.between(value, value + 24 * 3600 * 1000 - 1))
                    continue

                if operator == 'ne':
                    field = getattr(listing, name)
                    query = query.where(~field.between(value, value + 24 * 3600 * 1000 - 1))
                    continue

                if operator in ['le', 'gt']:
                    value = value + 24 * 3600 * 1000 - 1

            field = getattr(listing, name)
            if operator == 'lt':
                query = query.where(field < value)

            elif operator == 'le':
                query = query.where(field <= value)

            elif operator == 'eq':
                query = query.where(field == value)

            elif operator == 'ne':
                query = query.where(field != value)

            elif operator == 'ge':
                query = query.where(field >= value)

            elif operator == 'gt':
                query = query.where(field > value)

            elif operator == 'substring':
                query = query.where(escaped_contains(field, value))

            elif operator == 'startswith':
                query = query.where(escaped_startswith(field, value))

            elif operator == 'any':
                desc = findfirst(lambda x, name=name: x.key == name, descs)
                rel = Table(desc.table)
                through = Table(desc.through)
                rel_id = getattr(through, desc.rel_id)
                listing_id = getattr(through, desc.listing_id)
                query = query.where(listing.id.isin(Query.from_(through)
                                                    .join(rel)
                                                    .on(rel_id == rel.id)
                                                    .where(rel.value.isin(value))
                                                    .select(listing_id)
                                                    .distinct()))

            elif operator == 'all':
                desc = findfirst(lambda x, name=name: x.key == name, descs)
                rel = Table(desc.table)
                through = Table(desc.through)
                rel_id = getattr(through, desc.rel_id)
                listing_id = getattr(through, desc.listing_id)
                query = query.where(listing.id.isin(Query.from_(through)
                                                    .join(rel)
                                                    .on(rel_id == rel.id)
                                                    .where(rel.value.isin(value))
                                                    .select(listing_id)
                                                    .groupby(listing_id)
                                                    .having(fn.Count('*') == len(value))))

            elif operator == 'substring_any':
                desc = findfirst(lambda x, name=name: x.key == name, descs)
                rel = Table(desc.table)
                through = Table(desc.through)
                rel_id = getattr(through, desc.rel_id)
                listing_id = getattr(through, desc.listing_id)
                query = query.where(listing.id.isin(Query.from_(through)
                                                    .join(rel)
                                                    .on(rel_id == rel.id)
                                                    .where(escaped_contains(rel.value, value))
                                                    .select(listing_id)
                                                    .distinct()))

            elif operator == 'words':
                for word in value:
                    query = query.where(escaped_contains(field, word))
            else:
                raise UnknownOperator(operator)

        query = query.orderby(tag.value).groupby('id')
        return await txn.exq(query)

    @run_in_transaction
    async def delete_listing_rel_values_without_ref(self, descs, txn=None):
        for desc in [y for x in descs.values() for y in x if y.through]:
            await self._delete_values_without_ref(desc.table, desc.through, desc.rel_id, txn)

    @run_in_transaction
    async def query(self, collections=None, discarded=None, outdated=None, path=None, tags=None,
                    abbrev=None, missing=None, txn=None):
        # pylint: disable=too-many-arguments, too-many-locals

        abbrev = 10 if abbrev is True else abbrev
        discarded = bool(discarded)

        dataset = Table('dataset')
        query = Query.from_(dataset)

        if collections:
            collection = Table('collection')
            query = query.join(collection)\
                         .on(dataset.collection_id == collection.id)\
                         .where(collection.name.isin(collections))

        if outdated:
            bitmask = ValueWrapper(STATUS_OUTDATED)
            query = query.where(dataset.status.bitwiseand(bitmask).eq(bitmask))

        query = query.where(dataset.discarded == discarded)

        if path or missing:
            files = Table('file')
            if path:
                query = query.where(dataset.id.isin(Query.from_(files)
                                                    .where(escaped_startswith(files.path, path))
                                                    .select(files.dataset_id)
                                                    .distinct()))
            if missing:
                query = query.where(dataset.id.isin(Query.from_(files)
                                                    .where(files.missing == missing)
                                                    .select(files.dataset_id)
                                                    .distinct()))

        if tags:
            tag = Table('tag')
            dataset_tag = Table('dataset_tag')
            query = query.where(dataset.id.isin(Query.from_(dataset_tag)
                                                .join(tag)
                                                .on(dataset_tag.tag_id == tag.id)
                                                .where(tag.value.isin(tags))
                                                .select(dataset_tag.dataset_id)
                                                .distinct()))

        query = query.select(dataset.setid)\
                     .orderby(dataset.setid)
        datasets = await txn.exq(query)
        return [
            x['setid'][:abbrev] if abbrev else x['setid']
            for x in datasets
        ]

    @run_in_transaction
    async def update_listing_relations(self, desc, values, relations, txn=None):
        await self._ensure_values(desc.table, values, None, txn)
        await self._ensure_refs(desc.table, desc.through, desc.rel_id, desc.listing_id, relations,
                                '::', None, txn=txn)

        table, through = Tables(desc.table, desc.through)
        relid = getattr(through, desc.rel_id)
        backid = getattr(through, desc.listing_id)
        await txn.exq(Query.from_(through)
                      .where(backid.isin({x[1] for x in relations})
                             & ~Tuple(relid, backid).isin(
                                 self.get_resolve_value_query(table, relations, '::', None)))
                      .delete())

    @run_in_transaction  # noqa: C901
    async def rpc_query(self, model, filters, attrs, order, limit, offset,  # noqa: C901
                        user, txn=None):
        # pylint: disable=too-many-arguments, too-many-statements, too-many-locals, too-many-branches
        result = {}
        table = Table(model)
        try:
            tablemeta = [x for x in self.MODELS if x._meta.table == model][0]._meta
        except IndexError:
            raise ValueError(f'there is no model "{model}"')

        select = [
            getattr(table, k)
            for k, v in attrs.items()
            if v is True and k in tablemeta.db_fields
        ]
        if select:
            if 'id' not in attrs:
                select.append(table.id)
        else:
            select = [table.star]
        query = Query.from_(table).select(*select)

        for filt in filters:
            query = query.where(resolve_filter(table, filt, self.MODELS))

        def customize_query(query, model, table, through=None):
            if model in ['collection', 'dataset']:
                if through:
                    query = query.where(through.isin(self.get_actionable(model, user, 'list')))
                else:
                    query = query.where(table.id.isin(self.get_actionable(model, user, 'list')))
            elif model in ['group', 'user'] and not through:
                query = query.where(table.name.not_like('marv:%'))
            return query

        query = customize_query(query, model, table)

        if order:
            if not isinstance(order, list) or len(order) != 2:
                raise ValueError(f'Order "{order}" should be ["fieldname", "ASC|DESC"]')
            query = query.orderby(getattr(table, order[0]),
                                  order=Order.desc if order[1] == 'DESC' else Order.asc)
        if limit:
            query = query.limit(limit)
            if offset:
                query = query.offset(offset)

        qres = cleanup_attrs(await txn.exq(query))
        result.setdefault(model, []).extend(qres)

        ids = [x['id'] for x in qres]
        for fieldname in set(attrs.keys()) - set(tablemeta.db_fields):
            requested = attrs[fieldname]
            if requested is True:
                select = ['*']
            else:
                select = [k for k, v in requested.items() if v is True]
                if 'id' not in select:
                    select.append('id')

            if fieldname in tablemeta.backward_fk_fields:
                field = tablemeta.fields_map[fieldname]
                tablename = field.model_class._meta.table
                subtable = Table(tablename)
                if '*' not in select and field.relation_field not in select:
                    select.append(field.relation_field)
                query = Query.from_(subtable)\
                             .select(*select)\
                             .where(getattr(subtable, field.relation_field).isin(ids))
                query = customize_query(query, tablename, subtable)
                relres = await txn.exq(query)
                for prime in qres:
                    prime[fieldname] = [
                        x['id']
                        for x in relres
                        if x[field.relation_field] == prime['id']
                    ]
                result.setdefault(tablename, []).extend(cleanup_attrs(relres))

            elif fieldname in tablemeta.m2m_fields:
                field = tablemeta.fields_map[fieldname]
                through = Table(field.through)
                tablename = field.model_class._meta.table
                rel = Table(tablename)
                query = Query.from_(through)\
                             .select(field.backward_key, field.forward_key)\
                             .where(getattr(through, field.backward_key).isin(ids))
                query = customize_query(query, field.through, through,
                                        getattr(through, field.forward_key))
                refs = await txn.exq(query)
                query = Query.from_(rel)\
                             .select(*select)\
                             .where(rel.id.isin({x[field.forward_key] for x in refs}))
                query = customize_query(query, tablename, rel)
                relres = await txn.exq(query)
                relids = [x['id'] for x in relres]
                for prime in qres:
                    prime[fieldname] = [
                        x[field.forward_key]
                        for x in refs
                        if x[field.backward_key] == prime['id'] and x[field.forward_key] in relids
                    ]
                result.setdefault(tablename, []).extend(cleanup_attrs(relres))

            else:
                raise ValueError(f'field {fieldname} not on model {model}')

        return result

    async def restore_database(self, site, dct):
        async with scoped_session(self) as txn:
            for keys, handler in self.IMPORT_HANDLERS:
                if keys.issubset(dct.keys()):
                    await handler(site, dct, txn)
            if dct:
                log.warning(
                    'Some fields from the dump file could not be imported. If you are imporing '
                    'from a newer version of MARV of from a different MARV distribution (EE into '
                    'CE) this behavior is expected. Please be aware that some information from '
                    'the original dump was lost. The following fields were not processed:\n %s',
                    list(dct.keys()),
                )

    @classmethod
    async def dump_database(cls, dburi):
        """Dump database.

        The structure of the database is reflected and therefore also
        older databases, not matching the current version of marv, can be
        dumped.
        """

        class T2(Tortoise):
            apps = {}
            _connections = {}
            _inited = False

            @classmethod
            async def transaction(cls):
                await cls.init(db_url=dburi, modules={'models': []})
                connection = cls._connections['default']
                return connection._in_transaction()  # pylint: disable=protected-access

        dump = {}
        async with await T2.transaction() as txn:
            master = Table('sqlite_master')
            tables = {
                x['name']: Table(x['name'])
                for x in await get_items_for_query(
                    Query.from_(master)
                    .select(master.name)
                    .where((master.type == 'table')
                           & escaped_not_startswith(master.name, 'sqlite_')
                           & escaped_not_startswith(master.name, 'l_')),
                    txn,
                )
            }

            for tbls, dumper in cls.EXPORT_HANDLERS:
                assert tbls.issubset(tables.keys()), tbls
                await dumper(tables, dump, txn)
            assert not tables, tables.keys()

        await T2.close_connections()
        return dump



# Copyright 2016 - 2019  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from json import JSONDecodeError

from aiohttp import web

from .tooling import api_group as marv_api_group, generate_token as gentoken


@marv_api_group()
def auth(_):
    pass


@auth.endpoint('/auth', methods=['POST'], force_acl=['__unauthenticated__'])
async def auth_post(request):
    try:
        req = await request.json()
        username = req['username']
        password = req['password']
    except (JSONDecodeError, KeyError):
        raise web.HTTPBadRequest

    if not username or not password:
        raise web.HTTPUnprocessableEntity

    if not await request.app['site'].db.authenticate(username, password):
        raise web.HTTPUnprocessableEntity

    return web.json_response({
        'access_token': gentoken(username, request.app['config']['SECRET_KEY']).decode('utf-8'),
    })



# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

# pylint: disable=no-self-use

import unittest


class TestCase(unittest.TestCase):
    def test_(self):
        from . import pythonic_capnp  # pylint: disable=no-name-in-module,import-outside-toplevel

        # class Pythonic(Base):
        #     """foo"""
        #     schema = pythonic_capnp.Pythonic

        builder = pythonic_capnp.Pythonic.new_message()
        reader = builder.as_reader()
        assert reader



# Copyright 2016 - 2019  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

import json


async def test_collection(app, site, client):
    await site.db.bulk_tag([
        ('foo', 1),
        ('foo', 2),
    ], [], '::')

    res = await client.get('/marv/api/collection/notexist')
    assert res.status == 401

    res = await client.get_json('/marv/api/collection/hodge')
    assert len(res['listing']['widget']['data']['rows']) == 50

    res = await client.get('/marv/api/collection/hodge', params={
        'filter': json.dumps({'not exist': {'val': 42, 'op': 'eq'}}),
    })
    assert res.status == 400

    res = await client.get('/marv/api/collection/hodge', params={
        'filter': json.dumps({'name': {'val': 42, 'op': 'not exist'}}),
    })
    assert res.status == 400

    res = await client.get_json('/marv/api/collection/hodge')
    app['debug'] = True
    res2 = await client.get_json('/marv/api/collection/hodge')
    assert res == res2

    res = await client.get('/marv/api/collection/hodge', params={
        'filter': json.dumps({'name': {'val': 'filtér', 'op': 'eq'}}),
    })
    assert res.status == 200



# Copyright 2016 - 2019  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

import functools
import json
import os
import re
from collections import OrderedDict, defaultdict, namedtuple
from collections.abc import Mapping
from inspect import getmembers
from itertools import groupby
from logging import getLogger

from pypika import SQLLiteQuery as Query

from marv import utils
from marv.config import ConfigError, calltree, getdeps, make_funcs, parse_function
from marv.db import scoped_session
from marv.model import Collection as CollectionModel
from marv.model import Comment, Dataset, File
from marv.model import make_listing_model, make_table_descriptors
from marv_detail import FORMATTER_MAP, detail_to_dict
from marv_detail.types_capnp import Detail  # pylint: disable=no-name-in-module
from marv_node.node import Node
from marv_node.setid import SetID
from marv_store import Store


FILTER_OPERATORS = \
    'lt le eq ne ge gt substring startswith any all substring_any words'.split()
FILTER_MAP = {
    'datetime': lambda ns: None if ns is None else int(ns / 10**6),
    'filesize': lambda x: None if x is None else int(x),
    'float': lambda x: None if x is None else float(x),
    'int': lambda x: None if x is None else int(x),
    'subset': lambda lst: [str(x) for x in lst or []],
    'string': lambda x: None if x is None else str(x),
    'string[]': lambda lst: [str(x) for x in lst or []],
    'timedelta': lambda ns: None if ns is None else int(ns / 10**6),
    'words': lambda lst: ' '.join(lst or []),
}

Filter = namedtuple('Filter', 'name value operator type')
FilterSpec = namedtuple('FilterSpec', 'name title operators value_type function')
ListingColumn = namedtuple('ListingColumn', 'name heading formatter islist function')
SummaryItem = namedtuple('SummaryItem', 'id title formatter islist function')


def make_filter_spec(line):
    fields = [x.strip() for x in line.split('|', 4)]
    name, title, ops, value_type, function = fields
    ops = ops.split()
    return FilterSpec(name, title, ops, value_type, function)


def make_listing_column(line):
    fields = [x.strip() for x in line.split('|', 3)]
    name, heading, formatter, function = fields
    islist = formatter.endswith('[]')
    if islist:
        formatter = formatter[:-2]
    return ListingColumn(name, heading, formatter, islist, function)


def make_summary_item(line):
    fields = [x.strip() for x in line.split('|', 3)]
    name, title, formatter, function = fields
    islist = formatter.endswith('[]')
    if islist:
        formatter = formatter[:-2]
    return SummaryItem(name, title, formatter, islist, function)


def flatten_syntax(functree):
    functree = (flatten_syntax(x) if isinstance(x, tuple) else x for x in functree)
    functree = (x if len(x) > 1 else x[0] for x in functree)
    functree = ({} if x[1] == [] else
                ('map', {}, ('idx', x[1][0]), x[1][1]) if x[0] == 'rows' else
                x for x in functree)
    return tuple(functree)


def parse_summary(items):
    summary = []
    for item in items:
        functree, pos = parse_function(item.function)
        assert pos == len(item.function)
        dct = item._asdict()
        functree = flatten_syntax(functree)
        dct['function'] = functree
        summary.append(dct)
    return summary


def rowdumps(*args, **kw):
    return json.dumps(*args, sort_keys=True, separators=(',', ':'), **kw)


class Collections(Mapping):
    _collections = None

    @property
    def default_id(self):
        return next(iter(self._dct.keys()))

    @property
    def _dct(self):
        if self._collections:
            return self._collections
        self._collections = self.loadconfig(self.config, self.site)
        return self._collections

    @staticmethod
    def loadconfig(config, site):
        names = config.marv.collections
        collections = OrderedDict((x, Collection(config, x, site)) for x in names)
        assert not collections.keys() ^ names, (names, collections.keys())
        scanroots = [
            x for collection in collections.values()
            for x in collection.scanroots
        ]
        assert len(set(scanroots)) == len(scanroots),\
            'Scanroots must not be shared between collections'
        return collections

    def __init__(self, config, site):
        self.config = config
        self.site = site

    def __dir__(self):
        return list(self._dct.keys()
                    | self.__dict__.keys()
                    | {x[0] for x in getmembers(type(self))})

    def __getattr__(self, name):
        return self[name]

    def __getitem__(self, key):
        return self._dct[key]

    def __iter__(self):
        return iter(self._dct)

    def __len__(self):
        return len(self._dct)


def cached_property(func):
    """Create read-only property that caches its function's value."""
    @functools.wraps(func)
    def cached_func(self):
        cacheattr = f'_{func.__name__}'
        try:
            return getattr(self, cacheattr)
        except AttributeError:
            value = func(self)
            setattr(self, cacheattr, value)
            return value
    return property(cached_func)


class Collection:
    # pylint: disable=too-many-public-methods

    @property
    def compare(self):
        # TODO: should we load?
        return self.section.compare

    @cached_property
    def detail_deps(self):
        deps = set()
        deps.update(getdeps(self.detail_title))
        deps.update(self.section.detail_sections)
        deps.update(self.section.detail_summary_widgets)
        deps.difference_update(['comments', 'dataset', 'status', 'tags'])
        return deps

    @cached_property
    def detail_sections(self):
        nodes = self.nodes
        return [nodes[x] for x in self.section.detail_sections]

    @cached_property
    def detail_summary_widgets(self):
        nodes = self.nodes
        return [nodes[x] for x in self.section.detail_summary_widgets]

    @cached_property
    def detail_title(self):
        return self.section.detail_title

    @cached_property
    def filter_functions(self):
        funcs = []
        for spec in self.filter_specs.values():
            if spec.name in ('comments', 'status', 'tags'):
                continue
            functree, pos = parse_function(spec.function)
            assert pos == len(spec.function)
            funcs.append((spec, functree))
        return funcs

    @cached_property
    def filter_specs(self):
        specs = [make_filter_spec(line) for line in self.section.filters]
        names = [x.name for x in specs]
        assert all(re.match('^[a-z0-9_]+$', x) for x in names)
        assert 'setid' in names
        assert 'tags' in names
        return OrderedDict((x.name, x) for x in specs)

    @cached_property
    def listing_columns(self):
        return [make_listing_column(line) for line in self.section.listing_columns]

    @cached_property
    def listing_deps(self):
        deps = {x for y in self.filter_functions for x in getdeps(y[1])}
        deps.update(x for y in self.listing_functions for x in getdeps(y[1]))
        deps.difference_update(['comments', 'dataset', 'status', 'tags'])
        return deps

    @cached_property
    def listing_functions(self):
        funcs = []
        for col in self.listing_columns:
            functree, pos = parse_function(col.function)
            assert pos == len(col.function)
            funcs.append((col, functree))
        return funcs

    @cached_property
    def model(self):
        return make_listing_model(self.name, self.filter_specs)

    @cached_property
    def table_descriptors(self):
        return make_table_descriptors(self.model)

    @cached_property
    def nodes(self):
        nodes = OrderedDict()
        linemap = {}
        for line in self.section.nodes:
            try:
                nodename, node = utils.find_obj(line, True)
            except AttributeError:
                raise ConfigError(self.section, 'nodes', 'Cannot find node %s' % line)

            node = Node.from_dag_node(node)

            if node in linemap:
                raise ConfigError(self.section, 'nodes',
                                  '%s already listed as %s' %
                                  (line, linemap[node]))
            linemap[node] = line
            if nodename in nodes:
                raise ConfigError(self.section, 'nodes', 'duplicate name %s' % (nodename,))
            if not node.schema:
                raise ConfigError(self.section, 'nodes', '%s does not define a schema!' % (line,))
            nodes[nodename] = node
        return nodes

    @property
    def scanner(self):
        return utils.find_obj(self.section.scanner)

    @property
    def scanroots(self):
        return self.section.scanroots

    @cached_property
    def sortcolumn(self):
        listing_sort = self.section.listing_sort
        try:
            return 0 if not listing_sort[0] else \
                next(i for i, x in enumerate(self.listing_columns)
                     if x.name == listing_sort[0])
        except StopIteration:
            raise ValueError("No column named '%s'" % listing_sort[0])

    @cached_property
    def sortorder(self):
        listing_sort = self.section.listing_sort
        try:
            sortorder = listing_sort[1]
            orders = ['ascending', 'descending']
            return {x: x for x in orders}[sortorder]
        except IndexError:
            return 'ascending'
        except KeyError:
            raise ConfigError(self.section, 'listing_sort', f'{sortorder} not in {orders}')

    @cached_property
    def summary_items(self):
        items = [make_summary_item(line)
                 for line in self.section.listing_summary]
        return parse_summary(items)

    @property
    def section(self):
        return self.config[f'collection {self.name}']

    def __init__(self, config, name, site):
        self.config = config
        self.name = name
        self.site = site

    async def scan(self, scanpath, dry_run=False):  # noqa: C901
        # pylint: disable=too-many-locals,too-many-branches,too-many-statements

        log = getLogger('.'.join([__name__, self.name]))
        if not os.path.isdir(scanpath):
            log.warning('%s does not exist or is not a directory', scanpath)

        log.verbose("scanning %s'%s'", 'dry_run ' if dry_run else '', scanpath)

        # missing/changed flag for known files
        async with scoped_session(self.site.db) as connection:
            known_files = await File.filter(path__startswith=scanpath)\
                                    .filter(dataset__discarded__not=True)\
                                    .using_db(connection)
            known_filenames = defaultdict(set)
            changes = defaultdict(list)  # all mtime/missing changes in one transaction
            for file in known_files:
                path = file.path
                known_filenames[os.path.dirname(path)].add(os.path.basename(path))
                try:
                    mtime = utils.mtime(path)
                    missing = False
                except OSError:
                    mtime = None
                    missing = True
                if missing ^ bool(file.missing):
                    log.info("%s '%s'", 'lost' if missing else 'recovered', path)
                    changes[file.dataset_id].append((file, missing))
                if mtime and int(mtime * 1000) > file.mtime:
                    log.info("mtime newer '%s'", path)
                    changes[file.dataset_id].append((file, mtime))

            # Apply missing/mtime changes
            if not dry_run and changes:
                ids = changes.keys()
                for dataset in await Dataset.filter(id__in=ids).using_db(connection):
                    for file, change in changes.pop(dataset.id):
                        check_outdated = False
                        if isinstance(change, bool):
                            file.missing = change
                            dataset.missing = change
                        else:
                            file.mtime = int(change * 1000)
                            check_outdated = True
                        await file.save(connection)
                    if check_outdated:
                        await dataset.fetch_related('files', using_db=connection)
                        self._check_outdated(dataset)
                    dataset.time_updated = int(utils.now())
                    await dataset.save(connection)
                assert not changes

            # Scan for new files
            batch = []
            for directory, subdirs, filenames in utils.walk(scanpath):
                # Ignore directories containing a .marvignore file
                if os.path.exists(os.path.join(directory, '.marvignore')):
                    subdirs[:] = []
                    continue

                # Ignore hidden directories and traverse subdirs alphabetically
                subdirs[:] = sorted(x for x in subdirs if x[0] != '.')

                # Ignore hidden and known files
                known = known_filenames[directory]
                filenames = {x for x in filenames if x[0] != '.'}
                filenames = sorted(filenames - known)

                for name, files in self.scanner(directory, subdirs, filenames):
                    files = [x if os.path.isabs(x) else os.path.join(directory, x)
                             for x in files]
                    assert all(x.startswith(directory) for x in files), files
                    if dry_run:
                        log.info("would add '%s': '%s'", directory, name)
                    else:
                        dataset = await self.make_dataset(connection, files, name)
                        batch.append(dataset)
                        if len(batch) > 50:
                            await self._upsert_listing(connection, log, batch)

            if not dry_run and batch:
                await self._upsert_listing(connection, log, batch)

        log.verbose("finished %s'%s'", 'dry_run ' if dry_run else '', scanpath)

    def _check_outdated(self, dataset):
        storedir = self.config.marv.storedir
        setdir = os.path.join(storedir, str(dataset.setid))
        latest = [os.path.realpath(x)
                  for x in [os.path.join(setdir, x) for x in os.listdir(setdir)]
                  if os.path.islink(x)]
        oldest_mtime = utils.mtime(os.path.join(setdir, 'detail.json'))
        for nodedir in latest:
            for dirpath, _, filenames in utils.walk(nodedir):
                for name in filenames:
                    path = os.path.join(dirpath, name)
                    oldest_mtime = min(oldest_mtime, utils.mtime(path))
        dataset_mtime = max(x.mtime for x in dataset.files)
        dataset.outdated = int(oldest_mtime * 1000) < dataset_mtime

    async def restore_datasets(self, data, txn=None):
        log = getLogger('.'.join([__name__, self.name]))
        batch = []
        comments = []
        tags = []
        async with scoped_session(self.site.db, txn) as connection:
            for dataset in data:
                _comments = dataset.pop('comments')
                _tags = dataset.pop('tags')
                dataset = await self.make_dataset(connection, _restore=True, **dataset)
                comments.extend(Comment(dataset=dataset, **x) for x in _comments)
                tags.append((dataset, _tags))
                batch.append(dataset)
                if len(batch) > 50:
                    await Comment.bulk_create(comments, using_db=connection)
                    comments[:] = []
                    await self._upsert_listing(connection, log, batch)
                    await self._add_tags(connection, tags)
            await Comment.bulk_create(comments, using_db=connection)
            comments[:] = []
            await self._upsert_listing(connection, log, batch)
            await self._add_tags(connection, tags)

    async def _add_tags(self, connection, data):
        add = [(tag, dataset.id) for dataset, tags in data for tag in tags]
        await self.site.db.bulk_tag(add, [], '::', txn=connection)
        data[:] = []

    async def _upsert_listing(self, txn, log, batch, update=False):
        descs = self.table_descriptors
        rendered = [(dataset.id, *self.render_listing(dataset)) for dataset in batch]
        listing_values = ((id, rowdumps(row), *fields.values()) for id, row, fields, _ in rendered)
        relvalues = sorted((key, value, id)
                           for id, _, _, relfields in rendered
                           for key, values in relfields.items()
                           for value in values)

        await txn.execute_query(Query.into(descs[0].table)
                                .columns('id', 'row', *rendered[0][2].keys())
                                .insert(*listing_values)
                                .ignore()
                                .get_sql().replace('IGNORE', 'OR REPLACE'))

        for key, group in groupby(relvalues, key=lambda x: x[0]):
            group = list(group)
            values = {(x[1],) for x in group}
            relations = [(x[1], x[2]) for x in group]

            desc = [x for x in descs if x.key == key][0]
            await self.site.db.update_listing_relations(desc, values, relations, txn=txn)

        for dataset in batch:
            log.info(f'{"updated" if update else "added"} %r', dataset)
        batch[:] = []

    async def make_dataset(self, connection, files, name, time_added=None, discarded=False,
                           setid=None, status=0, timestamp=None, _restore=None):
        # pylint: disable=too-many-arguments
        time_added = int(utils.now() * 1000) if time_added is None else time_added

        collection = await CollectionModel.filter(name=self.name).using_db(connection).first()
        dataset = await Dataset.create(collection=collection,
                                       name=name,
                                       discarded=discarded,
                                       status=status,
                                       time_added=time_added,
                                       timestamp=0,
                                       setid=setid or SetID.random(),
                                       acn_id=collection.acn_id,
                                       using_db=connection)

        if _restore:
            files = [File(dataset=dataset, idx=i, **x) for i, x in enumerate(files)]
        else:
            files = [File(dataset=dataset, idx=i, mtime=int(utils.mtime(path) * 1000),
                          path=path, size=stat.st_size)
                     for i, (path, stat)
                     in enumerate((path, utils.stat(path)) for path in files)]

        dataset.timestamp = timestamp or max(x.mtime for x in files)
        await dataset.save(using_db=connection)
        await File.bulk_create(files, using_db=connection)

        await dataset.fetch_related('files', using_db=connection)

        storedir = self.config.marv.storedir
        store = Store(storedir, self.nodes)
        store.add_dataset(dataset, exists_okay=_restore)
        self.render_detail(dataset)
        return dataset

    def render_detail(self, dataset):
        storedir = self.config.marv.storedir
        setdir = os.path.join(storedir, str(dataset.setid))
        try:
            os.mkdir(setdir)
        except OSError:
            pass
        assert os.path.isdir(setdir), setdir
        store = Store(storedir, self.nodes)
        funcs = make_funcs(dataset, setdir, store)

        summary_widgets = [
            x[0]._reader for x in  # pylint: disable=protected-access
            [store.load(setdir, node, default=None) for node in self.detail_summary_widgets]
            if x
        ]

        sections = [
            x[0]._reader for x in  # pylint: disable=protected-access
            [store.load(setdir, node, default=None) for node in self.detail_sections]
            if x
        ]

        dct = {'title': calltree(self.detail_title, funcs),
               'sections': sections,
               'summary': {'widgets': summary_widgets}}
        detail = Detail.new_message(**dct).as_reader()
        dct = detail_to_dict(detail)
        fd = os.open(os.path.join(setdir, '.detail.json'),
                     os.O_CREAT | os.O_EXCL | os.O_WRONLY, 0o666)
        jsonfile = os.fdopen(fd, 'w')
        json.dump(dct, jsonfile, sort_keys=True)
        jsonfile.close()
        os.rename(os.path.join(setdir, '.detail.json'),
                  os.path.join(setdir, 'detail.json'))
        self._check_outdated(dataset)

    def render_listing(self, dataset):
        # pylint: disable=too-many-locals

        storedir = self.config.marv.storedir
        setdir = os.path.join(storedir, str(dataset.setid))
        store = Store(storedir, self.nodes)
        funcs = make_funcs(dataset, setdir, store)

        values = []
        for col, functree in self.listing_functions:
            value = calltree(functree, funcs)
            if value is not None:
                transform = FORMATTER_MAP[col.formatter + ('[]' if col.islist else '')]
                value = transform(value)
            values.append(value)
        row = {'id': dataset.id,
               'setid': str(dataset.setid),
               'tags': ['#TAGS#'],
               'values': values}
        fields = {}
        relfields = {}
        relations = [x.key for x in self.table_descriptors if x.key]
        for filter_spec, functree in self.filter_functions:
            value = calltree(functree, funcs)
            transform = FILTER_MAP[filter_spec.value_type]
            value = transform(value)
            target = relfields if filter_spec.name in relations else fields
            target[filter_spec.name] = value

        return row, fields, relfields

    async def update_listings(self, datasets, txn=None):
        assert datasets

        log = getLogger('.'.join([__name__, self.name]))
        async with scoped_session(self.site.db, txn) as txn:
            await self._upsert_listing(txn, log, datasets, update=True)



# Copyright 2016 - 2020  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

import json
from pathlib import Path

from aiohttp import web

from marv.db import DBPermissionError
from marv_node.setid import SetID
from .tooling import HTTPPermissionError
from .tooling import api_group as marv_api_group, get_local_granted, safejoin, sendfile


@marv_api_group()
def dataset(_):
    pass


@dataset.endpoint('/file-list', methods=['POST'], acl_key='download_raw')
async def file_list(request):
    try:
        ids = await request.json()
    except json.JSONDecodeError:
        raise web.HTTPBadRequest

    if not ids:
        raise web.HTTPBadRequest

    try:
        datasets = await request.app['site'].db.get_datasets_by_dbids(
            ids,
            user=request['username'],
            action='download_raw',
            prefetch=('files',),
        )
    except DBPermissionError:
        raise HTTPPermissionError(request)

    paths = sorted(x.path for dataset in datasets for x in dataset.files)
    urls = [f'dataset/{dataset.setid}/{x.idx}' for dataset in datasets for x in dataset.files]

    return web.json_response({'paths': paths, 'urls': urls})


async def _send_detail_json(request, setid, setdir):
    site = request.app['site']
    try:
        sets = await site.db.get_datasets_by_setids((setid,), user=request['username'],
                                                    action='read',
                                                    prefetch=('collections', 'comments', 'tags'))
        acl = await site.db.get_acl('dataset', sets[0].id, request['username'],
                                    get_local_granted(request))
    except DBPermissionError:
        raise HTTPPermissionError(request)

    try:
        with (setdir / 'detail.json').open() as f:
            detail = json.load(f)  # pylint: disable=redefined-outer-name
    except IOError:
        raise web.HTTPNotFound()

    dataset = sets[0]  # pylint: disable=redefined-outer-name
    detail.update({
        'acl': acl,
        'collection': dataset.collections[0].name,
        'id': dataset.id,
        'setid': str(dataset.setid),
        'comments': [{'author': x.author, 'text': x.text, 'timeAdded': x.time_added}
                     for x in sorted(dataset.comments, key=lambda x: x.time_added)],
        'tags': [x.value for x in sorted(dataset.tags, key=lambda x: x.value)],
        'all_known_tags':
            await site.db.get_all_known_tags_for_collection(dataset.collections[0].name),
    })

    resp = web.json_response(detail, headers={'Cache-Control': 'no-cache'})
    return resp


async def _get_filepath(request, setid, setdir, path):
    if path.isdigit():
        try:
            pathstr = await request.app['site'].db.get_filepath_by_setid_idx(
                setid,
                int(path),
                user=request['username'],
            )
        except DBPermissionError:
            raise HTTPPermissionError(request)
        path = Path(pathstr)

    else:
        try:
            await request.app['site'].db.get_datasets_by_setids([setid], [],
                                                                user=request['username'],
                                                                action='read')
        except DBPermissionError:
            raise HTTPPermissionError(request)
        path = safejoin(setdir, path)

    # Make sure path exists and is safe
    if not path.is_absolute() or not path.is_file():
        raise web.HTTPNotFound()
    return path


@dataset.endpoint('/dataset/{setid:[^/]+}{_:/?}{path:((?<=/).*)?}', acl_key='read')
async def detail(request):
    setid = request.match_info['setid']
    path = request.match_info['path'] or 'detail.json'
    try:
        setid = str(SetID(setid))
    except ValueError:
        raise web.HTTPBadRequest()

    setdir = Path(request.app['site'].config.marv.storedir) / setid

    if path == 'detail.json':
        return await _send_detail_json(request, setid, setdir)

    path = await _get_filepath(request, setid, setdir, path)

    return sendfile(path, request.path.split('/marv/api/dataset/')[0],
                    request.app['site'].config.marv.reverse_proxy)



# Copyright 2016 - 2019  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

import time
from collections import namedtuple

import pytest


async def try_listing(client):
    return await client.get('/marv/api/collection', headers=client.headers)


async def try_details(client):
    return await client.get('/marv/api/dataset/aaaaaaaaaaaaaaaaaaaaaaaaqa', headers=client.headers)


async def try_filelist(client):
    return await client.post('/marv/api/file-list', headers=client.headers, json=[1])


async def try_comment(client):
    return await client.post('/marv/api/comment', headers=client.headers, json={
        '1': {'add': ['lorem ipsum', 'dolor']},
    })


async def try_delete(client):
    return await client.delete('/marv/api/dataset', headers=client.headers, json=[1])


async def try_tag(client):
    return await client.post('/marv/api/tag', headers=client.headers, json={
        'hodge': {'add': {f'tag{time.time()}': [1]}},
    })


async def try_untag(client):
    tag = f'tag{time.time()}'

    auth = await client.unauthenticate()
    await client.authenticate('adm', 'adm_pw')
    await client.post('/marv/api/tag', headers=client.headers, json={
        'hodge': {'add': {tag: [1]}},
    })
    await client.unauthenticate()
    if auth:
        client.headers['Authorization'] = auth

    return await client.post('/marv/api/tag', headers=client.headers, json={
        'hodge': {'remove': {tag: [1]}},
    })


class AuthtestParams(namedtuple('Param', 'count expected forbidden local calls')):
    def __new__(cls, *args):
        assert len(args[-1]) == 7
        return super().__new__(cls, *args)


LOCAL_ACTIONS = {'comment', 'compare', 'delete', 'list', 'read', 'tag'}

UNAUTH_PUBLIC = AuthtestParams(
    2,
    {'auth_post'},
    {'rpc_entry'} | LOCAL_ACTIONS,
    {'download_raw', 'list', 'read'},
    [(200, try_listing),
     (200, try_details),
     (200, try_filelist),
     (401, try_comment),
     (401, try_tag),
     (401, try_untag),
     (401, try_delete)],
)

UNAUTH_AUTHENTICATED = AuthtestParams(
    0,
    {'auth_post'},
    {'rpc_entry'} | LOCAL_ACTIONS,
    set(),
    [(401, try_listing),
     (401, try_details),
     (401, try_filelist),
     (401, try_comment),
     (401, try_tag),
     (401, try_untag),
     (401, try_delete)],
)

AUTH = AuthtestParams(
    2,
    {'rpc_entry'},
    {'auth_post'} | LOCAL_ACTIONS,
    {'download_raw', 'list', 'read', 'comment', 'compare', 'tag'},
    [(200, try_listing),
     (200, try_details),
     (200, try_filelist),
     (200, try_comment),
     (200, try_tag),
     (200, try_untag),
     (403, try_delete)],
)

ADMIN = AuthtestParams(
    2,
    {'rpc_entry'},
    {'auth_post'} | LOCAL_ACTIONS,
    {'download_raw', 'list', 'read', 'comment', 'compare', 'tag', 'delete'},
    [(200, try_listing),
     (200, try_details),
     (200, try_filelist),
     (200, try_comment),
     (200, try_tag),
     (200, try_untag),
     (200, try_delete)],
)


@pytest.mark.parametrize('auth, params', [
    pytest.param(
        None, UNAUTH_PUBLIC,
        marks=pytest.mark.marv(site={'acl': 'marv_webapi.acls:public'}),
    ),
    pytest.param(
        ('test', 'test_pw'), AUTH,
        marks=pytest.mark.marv(site={'acl': 'marv_webapi.acls:public'}),
    ),
    pytest.param(
        ('adm', 'adm_pw'), ADMIN,
        marks=pytest.mark.marv(site={'acl': 'marv_webapi.acls:public'}),
    ),
    pytest.param(
        None, UNAUTH_AUTHENTICATED,
        marks=pytest.mark.marv(site={'acl': 'marv_webapi.acls:authenticated'}),
    ),
    pytest.param(
        ('test', 'test_pw'), AUTH,
        marks=pytest.mark.marv(site={'acl': 'marv_webapi.acls:authenticated'}),
    ),
    pytest.param(
        ('adm', 'adm_pw'), ADMIN,
        marks=pytest.mark.marv(site={'acl': 'marv_webapi.acls:authenticated'}),
    ),
])
async def test_profiles(client, auth, params):
    if auth:
        await client.authenticate(*auth)

    # check global permissions (non action routes) on meta
    res = await client.get_json('/marv/api/meta')
    assert not params.expected - set(res['acl'])
    assert not params.forbidden & set(res['acl'])
    assert len(res['collections']) == params.count

    # check local permissions on collection
    res = await client.get_json('/marv/api/collection')
    if params.local:
        assert set(res['acl']) == params.local
    else:
        assert res.status == 401

    # try accessing actions
    for code, func in params.calls:
        assert (await func(client)).status == code


async def test_authentication(client, site):
    res = await client.post('/marv/api/auth', json={})
    assert res.status == 400

    res = await client.post('/marv/api/auth', json={'username': 'foo', 'password': 'bar'})
    assert res.status == 422

    res = await client.post_json('/marv/api/auth', json={'username': 'test', 'password': 'test_pw'})
    assert res['access_token']

    await client.authenticate('test', 'test_pw')
    await site.db.user_rm('test')
    res = await client.get('/marv/api/meta', headers=client.headers)
    assert res.status == 401

    client.headers['Authorization'] = 'Basic XXXX'
    res = await client.get('/marv/api/meta', headers=client.headers)
    assert res.status == 401



# Copyright 2016 - 2019  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only


async def test_dataset(site, client):
    await client.authenticate('adm', 'adm_pw')

    sets = await site.db.get_datasets_for_collections(None)

    res = await client.delete('/marv/api/dataset', headers=client.headers, json=[])
    assert res.status == 400

    res = await client.delete('/marv/api/dataset', headers=client.headers, json=[1])
    assert res.status == 200

    rest = await site.db.get_datasets_for_collections(None)
    assert set(sets) - set(rest) == {sets[0]}

    await client.authenticate('test', 'test_pw')
    res = await client.delete('/marv/api/dataset', headers=client.headers, json=[2])
    assert res.status == 403



# Copyright 2016 - 2019  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from pathlib import Path

import pytest

from marv.tests.conftest import recorded
from marv.tests.test_dump_restore import postscan


DATADIR = Path(__file__).parent / 'data'


async def get_listings(client):
    metadata = await client.get_json('/marv/api/meta')
    listings = {}
    for colinfo in metadata['collections']:
        name = colinfo['name']
        listings[name] = await client.get_json(f'/marv/api/collection/{name}')
    return listings


def get_rows(listing):
    rows = listing['listing']['widget']['data']['rows']
    lst = []
    for row in sorted(rows, key=lambda x: x['setid']):
        del row['id']
        lst.append(row)
    return lst


@pytest.mark.marv(site={'empty': True})
async def test_empty_responses(client):
    assert recorded(await get_listings(client), DATADIR / 'empty_listings.json')


@pytest.mark.marv(site={'postscan': postscan, 'size': 2})
async def test_full_responses(client, site):
    sets = await site.db.get_datasets_for_collections(None)
    for setid in sets:
        await site.run(setid)

    listings = await get_listings(client)
    listings = {k: get_rows(v) for k, v in listings.items()}
    assert recorded(listings, DATADIR / 'full_listings.json')

    details = []
    for colname, rows in sorted(listings.items()):
        for row in rows:
            detail = await client.get_json(f'/marv/api/dataset/{row["setid"]}')
            del detail['id']
            assert detail['collection'] == colname
            details.append(detail)
    assert recorded(details, DATADIR / 'full_details.json')



# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

import os
import re
import sys
import time
from datetime import datetime, timedelta
from datetime import tzinfo as tzinfo_base
from itertools import islice
from subprocess import Popen as _Popen

from marv_api.utils import NOTSET, find_obj  # pylint: disable=unused-import
from marv_node.setid import decode_setid, encode_setid  # pylint: disable=unused-import


def echo(*args, **kw):
    """Wrap print to let linter forbid print usage."""
    print(*args, **kw)  # noqa: T001


def err(*args, exit=None, **kw):
    """Print to stderr and optionally exit."""
    print(*args, **kw, file=sys.stderr, flush=True)  # noqa: T001
    if exit is not None:
        sys.exit(exit)


def chunked(iterable, chunk_size):
    itr = iter(iterable)
    return iter(lambda: tuple(islice(itr, chunk_size)), ())


def findfirst(predicate, iterable, default=NOTSET):
    try:
        return next(x for x in iterable if predicate(x))
    except StopIteration:
        if default is not NOTSET:
            return default
        raise ValueError('No item matched predicate!')


def mtime(path):
    """Wrap os.stat() st_mtime for ease of mocking."""
    return os.stat(path).st_mtime


def stat(path):
    """Wrap os.stat() for ease of mocking."""  # noqa: D402
    # TODO: https://github.com/PyCQA/pydocstyle/issues/284
    return os.stat(path)


def walk(path):
    """Wrap os.walk() for ease of mocking."""  # noqa: D402
    # TODO: https://github.com/PyCQA/pydocstyle/issues/284
    return os.walk(path)


def now():
    """Wrap time.time() for ease of mocking."""
    return time.time()


def parse_filesize(string):
    val, unit = re.match(r'^\s*([0-9.]+)\s*([kmgtpezy]b?)?\s*$', string, re.I)\
                  .groups()
    val = float(val)
    if unit:
        val *= 1 << (10 * (1 + 'kmgtpezy'.index(unit.lower()[0])))
    return int(val)


def parse_datetime(string):
    class TZInfo(tzinfo_base):
        def __init__(self, offset=None):
            self.offset = offset

        def dst(self, dt):
            raise NotImplementedError()

        def tzname(self, dt):
            return self.offset

        def utcoffset(self, dt):
            if self.offset == 'Z':
                hours, minutes = 0, 0
            else:
                hours, minutes = self.offset[1:].split(':')
            offset = timedelta(hours=int(hours), minutes=int(minutes))
            return offset if self.offset[0] == '+' else -offset

    groups = re.match(r'^(\d\d\d\d)-(\d\d)-(\d\d)T'
                      r'(\d\d):(\d\d):(\d\d)((?:[+-]\d\d:\d\d)|Z)$', string)\
               .groups()
    tzinfo = TZInfo(groups[-1])
    return datetime(*(int(x) for x in groups[:-1]), tzinfo=tzinfo)


def parse_timedelta(delta):
    match = re.match(r'^\s*(?:(\d+)\s*h)?'
                     r'\s*(?:(\d+)\s*m)?'
                     r'\s*(?:(\d+)\s*s?)?\s*$', delta)
    h, m, s = match.groups() if match else (None, None, None)  # pylint: disable=invalid-name
    return (int(h or 0) * 3600 + int(m or 0) * 60 + int(s or 0)) * 1000


def profile(func, sort='cumtime'):
    # pylint: disable=import-outside-toplevel
    import functools
    import pstats
    from cProfile import Profile
    _profile = Profile()

    @functools.wraps(func)
    def profiled(*args, **kw):
        _profile.enable()
        result = func(*args, **kw)
        _profile.disable()
        stats = pstats.Stats(_profile).sort_stats(sort)
        stats.print_stats()
        return result
    return profiled


def underscore_to_camelCase(string):  # pylint: disable=invalid-name
    return ''.join(x.capitalize() for x in string.split('_'))


def sanitize_env(env):
    ld_library_path = env.get('LD_LIBRARY_PATH')
    if ld_library_path:
        clean_path = ':'.join([
            x
            for x in ld_library_path.split(':')
            if not x.startswith('/tmp/_MEI')
        ])
        if clean_path:
            env['LD_LIBRARY_PATH'] = clean_path
        else:
            del env['LD_LIBRARY_PATH']
    return env


def popen(*args, env=None, **kw):
    env = sanitize_env(os.environ.copy() if env is None else env)
    return _Popen(*args, env=env, **kw)


def within_pyinstaller_bundle():
    return any(x for x in sys.path if '/tmp/_MEI' in x)


def within_staticx_bundle():
    return bool(os.environ.get('STATICX_PROG_PATH'))



# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from ..testing import make_dataset, marv, run_nodes


@marv.node(group='ondemand')
def source():
    out = yield marv.create_stream('foo')
    yield out.msg(1)


@marv.node()
@marv.input('handle', default=marv.select(source, 'foo'))
def consumer(handle):
    while True:
        msg = yield marv.pull(handle)
        if msg is None:
            break
        yield marv.push(msg)


DATASET = make_dataset()


async def test():
    nodes = [consumer]
    streams = await run_nodes(DATASET, nodes)
    assert streams == [
        [1],
    ]



# Copyright 2016 - 2020  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

import logging

from testfixtures import LogCapture

from ..testing import make_dataset, marv, run_nodes

DATASET = make_dataset()
SETID = DATASET.setid


@marv.node(group=True)
def multi():
    logger = yield marv.get_logger()
    logger.critical('multi started')
    a_out = yield marv.create_stream('a')
    b_out = yield marv.create_stream('b')
    assert a_out.name == 'a'
    assert b_out.name == 'b'

    yield a_out.msg(1)
    yield b_out.msg(10)
    yield a_out.msg(2)
    yield b_out.msg(20)
    logger.critical('multi finished')


@marv.node()
@marv.input('node', default=multi)
def consumer(node):
    logger = yield marv.get_logger()
    logger.critical('consumer started')
    a_in, b_in = yield marv.pull_all(node, node)
    assert a_in.name == 'a'
    assert b_in.name == 'b'
    # streams = yield marv.io(*node)  # like for values
    # streams = yield marv.io(node, consume=True)  # a bit less voodoo

    values = []
    while True:
        msgs = yield marv.pull_all(a_in, b_in)
        if msgs == [None, None]:
            break
        values.append(msgs)
    assert values == [[1, 10], [2, 20]]
    logger.critical('consumer finished')


async def test():
    nodes = [consumer]
    with LogCapture(level=logging.CRITICAL) as log:
        await run_nodes(DATASET, nodes)

    assert [x.msg for x in log.records] == [
        'multi started',
        'consumer started',
        'multi finished',
        'consumer finished',
    ]



# Copyright 2016 - 2020  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

import logging
from itertools import product

from testfixtures import LogCapture

from ..testing import make_dataset, marv, run_nodes

DATASET = make_dataset()
SETID = DATASET.setid


@marv.node(group='ondemand')
def source():
    logger = yield marv.get_logger()
    requested = yield marv.get_requested()
    logger.critical([x.name for x in requested])
    creates = [marv.create_stream(x.name) for x in requested]
    streams = []
    for create in creates:
        stream = yield create
        streams.append(stream)
    msgs = list(product(streams, [1, 2]))
    for stream, msg in msgs:
        yield marv.push(stream.msg(msg))


@marv.node()
@marv.input('stream', default=marv.select(source, 'a'))
def stream_a(stream):
    while True:
        msg = yield marv.pull(stream)
        if msg is None:
            break
        msg = yield marv.push(msg)


@marv.node()
@marv.input('stream', default=marv.select(source, 'b'))
def stream_b(stream):
    while True:
        msg = yield marv.pull(stream)
        if msg is None:
            break
        msg = yield marv.push(msg)


@marv.node(group=True)
@marv.input('stream1', default=stream_a)
@marv.input('stream2', default=stream_b)
def merged(stream1, stream2):
    yield marv.push(stream1)
    yield marv.push(stream2)


@marv.node()
@marv.input('stream', foreach=merged)
def consumer(stream):
    logger = yield marv.get_logger()
    while True:
        msg = yield marv.pull(stream)
        if msg is None:
            break
        logger.critical((stream.node.name, stream.name, msg))


async def test():
    nodes = [consumer]
    await run_nodes(DATASET, nodes)

    with LogCapture(level=logging.CRITICAL) as log:
        await run_nodes(DATASET, nodes)

    assert [x.msg for x in log.records] == [
        ['b'],
        ['a', 'b'],
        ('stream_a', 'default', 1),
        ('stream_b', 'default', 1),
        ('stream_a', 'default', 2),
        ('stream_b', 'default', 2),
    ]



# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

import time

from aiohttp import web

from marv.db import DBPermissionError
from .tooling import HTTPPermissionError, api_endpoint as marv_api_endpoint


@marv_api_endpoint('/comment', methods=['POST'])
async def comment(request):
    username = request['username']
    changes = await request.json()
    if not changes:
        raise web.HTTPBadRequest()

    now = int(time.time() * 1000)
    comments = [
        {'dataset_id': id, 'author': username, 'time_added': now, 'text': text}
        for id, ops in changes.items()
        for text in ops.get('add', [])
    ]
    try:
        await request.app['site'].db.bulk_comment(comments, user=username)
    except DBPermissionError:
        raise HTTPPermissionError(request)

    return web.json_response({})



# Copyright 2016 - 2019  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

import logging
import sqlite3
from collections import OrderedDict, namedtuple

import tortoise
from tortoise.fields import (
    BooleanField,
    CharField,
    DatetimeField,
    FloatField,
    ForeignKeyField,
    IntField,
    ManyToManyField,
    RESTRICT,
    TextField,
)
from tortoise.models import Model

from . import model_fields as custom
from .utils import underscore_to_camelCase

STATUS = OrderedDict((
    ('error', 'ERROR: One or more errors occured while processing this dataset'),
    ('missing', 'MISSING: One or more files of this dataset are missing'),
    ('outdated', 'OUTDATED: There are outdated nodes for this dataset'),
    ('pending', 'PENDING: There are pending node runs'),
))
STATUS_ERROR = 1
STATUS_MISSING = 2
STATUS_OUTDATED = 4
STATUS_PENDING = 8


tortoise.logger.setLevel(logging.WARN)


def patch_pypika_boolean_for_old_sqlite():
    from pypika.terms import ValueWrapper  # pylint: disable=import-outside-toplevel

    _original = ValueWrapper.get_value_sql

    def get_value_sql(self, **kwargs):
        value = _original(self, **kwargs)
        return {'true': 1, 'false': 0}.get(value, value)

    ValueWrapper.get_value_sql = get_value_sql


if sqlite3.sqlite_version_info < (3, 23, 0):
    patch_pypika_boolean_for_old_sqlite()


def make_status_property(bitmask, doc=None):
    def fget(obj):
        return obj.status & bitmask

    def fset(obj, value):
        status = obj.status or 0
        if value:
            obj.status = status | bitmask
        else:
            obj.status = status - (status & bitmask)

    def fdel(obj):
        fset(obj, False)

    return property(fget, fset, fdel, doc)


class Acn(Model):
    id = IntField(pk=True)


class Collection(Model):
    id = IntField(pk=True)
    name = CharField(max_length=32, unique=True)
    acn = ForeignKeyField('models.Acn', related_name='collections', on_delete=RESTRICT)


class Dataset(Model):
    id = IntField(pk=True)

    collection = ForeignKeyField('models.Collection', related_name='datasets')
    discarded = BooleanField(default=False)
    name = TextField()
    status = IntField(default=0)
    time_added = IntField()  # ms since epoch
    timestamp = IntField()  # ms since epoch

    setid = custom.SetIDField(unique=True)
    tags = ManyToManyField('models.Tag', related_name='datasets')
    acn = ForeignKeyField('models.Acn', related_name='datasets', on_delete=RESTRICT)

    # Populated by backreferences: comments, files

    error = make_status_property(1)
    missing = make_status_property(2)
    outdated = make_status_property(4)
    pending = make_status_property(8)

    def __repr__(self):
        return f'<{type(self).__name__} {self.setid} {self.name}>'


class File(Model):
    id = IntField(pk=True)
    dataset = ForeignKeyField('models.Dataset', related_name='files')
    idx = IntField()  # files are counted per dataset

    missing = BooleanField(default=False)
    mtime = IntField()  # ms since epoch
    path = TextField()
    size = IntField()

    def __repr__(self):
        return f"<{type(self).__name__} '{self.path}'>"


class Metadata(Model):
    id = IntField(pk=True)
    key = CharField(max_length=32, unique=True)
    value = TextField()


class Comment(Model):
    id = IntField(pk=True)
    dataset = ForeignKeyField('models.Dataset', related_name='comments')

    author = TextField()
    time_added = IntField()  # ms since epoch
    text = TextField()


class Tag(Model):
    id = IntField(pk=True)
    value = CharField(max_length=255, unique=True)


class User(Model):
    id = IntField(pk=True)
    name = CharField(max_length=255, unique=True)
    password = TextField(null=True)
    given_name = TextField(null=True)
    family_name = TextField(null=True)
    email = TextField(null=True)
    realm = TextField()
    realmuid = TextField(null=True)
    time_created = DatetimeField(auto_now_add=True)
    time_updated = DatetimeField(auto_now=True)
    groups = ManyToManyField('models.Group', related_name='users')


class Group(Model):
    id = IntField(pk=True)
    name = CharField(max_length=255, unique=True)


class Leaf(Model):
    id = IntField(pk=True)
    name = CharField(max_length=255, unique=True)
    access_token = TextField()
    refresh_token = TextField()
    time_created = DatetimeField(auto_now_add=True)
    time_updated = DatetimeField(auto_now=True)


__models__ = [
    Acn,
    Collection,
    Comment,
    Dataset,
    File,
    Group,
    Leaf,
    Metadata,
    Tag,
    User,
]


ListingModel = namedtuple('ListingModel', 'Listing relations secondaries')
ListingDescriptor = namedtuple('ListingDescriptor', 'table key through rel_id listing_id')


def make_listing_model(name, filter_specs):
    models = []
    listing_name = f'l_{name}'
    listing_model_name = underscore_to_camelCase(listing_name)

    def generate_relation(fname):
        rel_name = f'{listing_name}_{fname}'
        rel_model_name = underscore_to_camelCase(rel_name)
        assert '_' not in rel_model_name, rel_model_name

        model = type(rel_model_name, (Model,), {
            'Meta': type('Meta', (), {'table': rel_name}),
            'id': IntField(pk=True),
            'value': CharField(max_length=255, index=True, unique=True),
            '__repr__': lambda self: f'<{type(self).__name__} {self.id} {self.value!r}>',
        })

        models.append(model)
        return ManyToManyField(f'models.{rel_model_name}', related_name='listing')

    coltype_factories = {
        # All dates and times in ms (since epoch)
        'datetime': lambda name: IntField(null=True),
        'filesize': lambda name: IntField(null=True),
        'float': lambda name: FloatField(null=True),
        'int': lambda name: IntField(null=True),
        'string': lambda name: TextField(null=True),
        'string[]': generate_relation,
        'subset': generate_relation,
        'timedelta': lambda name: IntField(null=True),
        'words': lambda name: TextField(null=True),
    }

    dct = {
        'Meta': type('Meta', (), {'table': listing_name}),
        'id': IntField(pk=True),
        'row': TextField(null=True),
        '__repr__': lambda self: f'<{type(self).__name__} {self.dataset_id}>',
    }

    dct.update((fspec.name, coltype_factories[fspec.value_type](fspec.name))
               for fspec in filter_specs.values()
               if fspec.name not in ('dataset', 'row', 'comments', 'status', 'tags'))

    models.append(type(listing_model_name, (Model,), dct))
    return models


def make_table_descriptors(models):
    listing = models[-1]
    meta = listing._meta

    result = [ListingDescriptor(meta.table, '', '', '', '')]
    for key in meta.m2m_fields:
        field = meta.fields_map[key]
        table = field.model_class.Meta.table
        result.append(
            ListingDescriptor(table, key, field.through, field.forward_key, field.backward_key))

    return result



# Copyright 2016 - 2020  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

# pylint: disable=invalid-name

import pytest

from ..testing import make_dataset, marv, run_nodes


class FALSISH:
    def __bool__(self):
        return False

    def __repr__(self):
        return '<FALSISH>'


FALSISH = FALSISH()


class DONTBOOLME:
    def __bool__(self):
        raise Exception

    def __repr__(self):
        return '<DONTBOOLME>'


DONTBOOLME = DONTBOOLME()


@marv.node()
def source():
    # yield marv.push(None) -- So far, this means we're done
    yield marv.push(0)
    yield marv.push(0.0)
    yield marv.push(False)
    yield marv.push('')
    yield marv.push(FALSISH)
    yield marv.push(DONTBOOLME)


@marv.node()
@marv.input('stream', default=source)
def consumer1(stream):
    yield marv.push(42)
    while True:
        msg = yield marv.pull(stream)
        if msg is None:
            break
        yield marv.push(msg)


@marv.node()
@marv.input('stream', default=source)
def consumer2(stream):
    yield marv.push(42)
    while True:
        msg, = yield marv.pull_all(stream)
        if msg is None:
            break
        yield marv.push(msg)


DATASET = make_dataset()


async def test():
    with pytest.raises(Exception):
        if DONTBOOLME:
            pass

    nodes = [source, consumer1, consumer2]
    streams = await run_nodes(DATASET, nodes)
    assert streams == [
        [0, 0.0, False, '', FALSISH, DONTBOOLME],
        [42, 0, 0.0, False, '', FALSISH, DONTBOOLME],
        [42, 0, 0.0, False, '', FALSISH, DONTBOOLME],
    ]
